//-----------------------------------------------------------------------------
//
// RTextView.cpp
//
// Reusable text view class.
//
// Adds following features to BTextView:
//
// Changes to ibeam cursor when over view.
// Handles home, end, and page up/down keys.
// Fixes Cut and Paste to scroll to selection.
// Fixes Cut and Paste so they do nothing if !IsEditable().
// Adds some useful scroll selection to functions.
//
//-----------------------------------------------------------------------------

#include "RTextView.h"
#include "RGlobals.h"

B_DEFINE_CLASS_INFO(RTextView, BTextView);

///RTextView
//-----------------------------------------------------------------------------

RTextView::RTextView(BRect frame, const char *name, BRect textRect, 
	ulong resizingMode, ulong flags)
	: BTextView(frame, name, textRect, resizingMode, flags)
{
}

///MouseMoved
//-----------------------------------------------------------------------------

void RTextView::MouseMoved(BPoint point, ulong transit, BMessage *msg)
{
	if (Window()->IsActive()) {
		if (transit == B_ENTERED_VIEW) {
			be_app->SetCursor(IBEAM_CURSOR);
		} else if (transit == B_EXITED_VIEW) {
			be_app->SetCursor(B_HAND_CURSOR);
		}
	}
	inherited::MouseMoved(point, transit, msg);
}

///WindowActivated
//-----------------------------------------------------------------------------

void RTextView::WindowActivated(bool active)
{
	if (active) {
		BPoint cursor;
		ulong buttons;
		GetMouse(&cursor, &buttons, FALSE);
		if (Bounds().Contains(cursor)) be_app->SetCursor(IBEAM_CURSOR);
	} else {
		be_app->SetCursor(B_HAND_CURSOR);
	}
	inherited::WindowActivated(active);
}

///Cut
//-----------------------------------------------------------------------------

void RTextView::Cut(BClipboard *clipboard)
{
	if (!IsEditable()) return;
	inherited::Cut(clipboard);
	ScrollToSelection();
}

///Paste
//-----------------------------------------------------------------------------

void RTextView::Paste(BClipboard *clipboard)
{
	if (!IsEditable()) return;
	inherited::Paste(clipboard);
	ScrollToSelection();
}

///Draw
//-----------------------------------------------------------------------------

void RTextView::Draw(BRect updateRect)
{
	// Need to check for set to ibeam on draws because of menus.
	// e.g., pull down menu from main menu bar, cursor changes to
	// hand, select menu command which is "on top of" text view,
	// release mouse without moving it - cursor is still the hand!
	
	inherited::Draw(updateRect);
	if (Window()->IsActive()) {
		BPoint cursor;
		ulong buttons;
		GetMouse(&cursor, &buttons, FALSE);
		if (updateRect.Contains(cursor)) be_app->SetCursor(IBEAM_CURSOR);
	}
}

///KeyDown
//-----------------------------------------------------------------------------

void RTextView::KeyDown(ulong aChar)
{
	if (aChar == B_HOME || aChar == B_END || aChar == B_PAGE_UP || 
		aChar == B_PAGE_DOWN) 
	{
		BScrollView *scrollView = cast_as(Parent(), BScrollView);
		if (!scrollView) return;
		BScrollBar *vScroll = scrollView->ScrollBar(B_VERTICAL);
		if (!vScroll) return;
		long min, max;
		vScroll->GetRange(&min, &max);
		long smallStep, bigStep;
		vScroll->GetSteps(&smallStep, &bigStep);
		long value = vScroll->Value();
		switch (aChar) {
			case B_HOME:
				Select(0, 0);
				value = min;
				break;
			case B_END:
				Select(TextLength(), TextLength());
				value = max - bigStep/2;
				break;
			case B_PAGE_UP:
				value -= bigStep;
				break;
			case B_PAGE_DOWN:
				value += bigStep;
				break;
		}
		vScroll->SetValue(value);
	} else {
		inherited::KeyDown(aChar);
	}
}	

///ScrollSelectionTo
//-----------------------------------------------------------------------------

void RTextView::ScrollSelectionTo(float x, bool evenIfAlreadyVisible)
{
	BScrollView *scrollView = cast_as(Parent(), BScrollView);
	if (scrollView) {
		BScrollBar *vScroll = scrollView->ScrollBar(B_VERTICAL);
		if (vScroll) {
//			if (evenIfAlreadyVisible) {
//				long min, max;
//				vScroll->GetRange(&min, &max);
//				vScroll->SetValue(max);
//			}
			BRect bounds = Bounds();
			long firstVis = IndexAtPoint(bounds.left, bounds.top);
			long lastVis = IndexAtPoint(bounds.right, bounds.bottom);
			ScrollToSelection();
			long start, finish;
			GetSelection(&start, &finish);
			long viewHeight = bounds.bottom - bounds.top;
			if (start <= firstVis && x != 0.0) {
				vScroll->SetValue(vScroll->Value() - viewHeight*x);
			} else if (lastVis <= finish && x != 1.0) {
				long smallStep, largeStep;
				vScroll->GetSteps(&smallStep, &largeStep);
				vScroll->SetValue(vScroll->Value() + 
					viewHeight*(1.0-x) - smallStep);
			}
			if (x != 0.0) {
				long val = vScroll->Value();
				long min, max;
				vScroll->GetRange(&min, &max);
				if (val == max) vScroll->SetValue(val - viewHeight*x);
			}
		}
	}
}

///ScrollSelectionToMiddle
//-----------------------------------------------------------------------------

void RTextView::ScrollSelectionToMiddle(bool evenIfAlreadyVisible)
{
	ScrollSelectionTo(0.5, evenIfAlreadyVisible);
}

///ScrollSelectionToTop
//-----------------------------------------------------------------------------

void RTextView::ScrollSelectionToTop(bool evenIfAlreadyVisible)
{
	ScrollSelectionTo(0.0, evenIfAlreadyVisible);
}
