//-----------------------------------------------------------------------------
//
// RWindowsMenu.cpp
//
// Reusable class for maintaining Windows menus in windows. This object
// runs in its own blooper thread to avoid deadlocks.
//
//-----------------------------------------------------------------------------

#include "RWindowsMenu.h"
#include "RWindow.h"

B_DEFINE_CLASS_INFO(RWindowsMenu, BLooper);

///RWindowsMenu
//-----------------------------------------------------------------------------

RWindowsMenu::RWindowsMenu(const char *menuBarViewName, 
	const char *windowsMenuName,
	const ulong windowsMenuMessageWhat,
	const char *windowsMenuMessageObjName)
		: BLooper(B_NORMAL_PRIORITY),
		menuBarViewName(menuBarViewName), 
		windowsMenuName(windowsMenuName),
		windowsMenuMessageWhat(windowsMenuMessageWhat),
		windowsMenuMessageObjName(windowsMenuMessageObjName)
{
	Run();
}

///Init
//-----------------------------------------------------------------------------

void RWindowsMenu::Init(BWindow *wind)
{
	// Initialize the Windows menu for a new window. Works by
	// finding some other window and cloning its Windows menu.

	BMessage *msg = new BMessage('init');
	msg->AddObject(windowsMenuMessageObjName, wind);
	PostMessage(msg);
}

///Add
//-----------------------------------------------------------------------------

void RWindowsMenu::Add(BWindow *wind, bool check)
{
	// Add a new window to the Windows menu in all open windows.

	BMessage *msg = new BMessage('add ');
	msg->AddObject(windowsMenuMessageObjName, wind);
	msg->AddBool("check", check);
	PostMessage(msg);
}

///Remove
//-----------------------------------------------------------------------------

void RWindowsMenu::Remove(BWindow *wind)
{
	// Remove a window from the the Windows menu in all open windows.

	BMessage *msg = new BMessage('rmve');
	msg->AddObject(windowsMenuMessageObjName, wind);
	PostMessage(msg);
}

///Check
//-----------------------------------------------------------------------------

void RWindowsMenu::Check(BWindow *wind, bool check)
{
	// Change the check mark state for a window in the Windows menu
	// in all open windows.

	BMessage *msg = new BMessage('chck');
	msg->AddObject(windowsMenuMessageObjName, wind);
	msg->AddBool("check", check);
	PostMessage(msg);
}

///MessageReceived
//-----------------------------------------------------------------------------

void RWindowsMenu::MessageReceived(BMessage *msg)
{
	BWindow *wind = (BWindow*)msg->FindObject(windowsMenuMessageObjName);
	bool check = msg->FindBool("check");
	ulong what = msg->what;
	bool done = FALSE;
	
	BList windowList = RWindow::WindowList();
	for (long i = 0; 
		RWindow *w = (RWindow*)windowList.ItemAt(i); 
		i++) 
	{
		if (w->Lock()) {
			BMenuBar *b = (BMenuBar*)w->FindView(menuBarViewName);
			if (b) {
				BMenuItem *x = (BMenuItem*)b->FindItem(windowsMenuName);
				if (x) {
					BMenu *m = x->Submenu();
					if (m) {
						if (what == 'init') {
							if (w != wind) {
								CloneMenu(wind, m);
								done = TRUE;
							}
						} else {
							AdjustOneMenu(m, what, wind, check);
						}
					}
				}
			}
			w->Unlock();
		}
		if (done) break;
	}
}

///CloneMenu
//-----------------------------------------------------------------------------

void RWindowsMenu::CloneMenu(BWindow *newWind, BMenu *oldMenu)
{
	if (newWind->Lock()) {
		BMenuBar *newBar = (BMenuBar*)newWind->FindView(menuBarViewName);
		if (newBar) {
			BMenuItem *x = (BMenuItem*)newBar->FindItem(windowsMenuName);
			if (x) {
				BMenu *newMenu = x->Submenu();
				if (newMenu) {
					long count = oldMenu->CountItems();
					for (long i = 0; i < count; i++) {
						BMenuItem *oldItem = oldMenu->ItemAt(i);
						BMessage *oldMsg = oldItem->Message();
						BMessage *newMsg = new BMessage(windowsMenuMessageWhat);
						newMsg->AddObject(windowsMenuMessageObjName, 
							oldMsg->FindObject(windowsMenuMessageObjName));
						BMenuItem *newItem = 
							new BMenuItem(oldItem->Label(), newMsg);
						newItem->SetMarked(oldItem->IsMarked());
						newItem->SetTarget(be_app);
						newMenu->AddItem(newItem);
					}
				}
			}
		}
		newWind->Unlock();
	}
}

///AdjustOneMenu
//-----------------------------------------------------------------------------

void RWindowsMenu::AdjustOneMenu(BMenu *menu, ulong what, BWindow *wind,
	bool check)
{
	long count = menu->CountItems();
	
	switch (what) {
		case 'add ':
			for (long i = 0; i < count; i++) {
				BMenuItem *item = menu->ItemAt(i);
				BMessage *msg = item->Message();
				BWindow *w =
					(BWindow*)msg->FindObject(windowsMenuMessageObjName);
				if (w == wind) return;
			}
			while (count--) {
				BMenuItem *item = menu->ItemAt(count);
				if (strcmp(wind->Title(), item->Label()) > 0) break;
			}
			BMessage *msg = new BMessage(windowsMenuMessageWhat);
			msg->AddObject(windowsMenuMessageObjName, wind);
			BMenuItem *item = new BMenuItem(wind->Title(), msg);
			item->SetMarked(check);
			item->SetTarget(be_app);
			menu->AddItem(item, count+1);
			break;
		case 'rmve':
			while (count--) {
				BMenuItem *item = menu->ItemAt(count);
				BMessage *msg = item->Message();
				BWindow *w =
					(BWindow*)msg->FindObject(windowsMenuMessageObjName);
				if (w == wind){
					menu->RemoveItem(item);
					break;
				}
			}
			break;
		case 'chck':
			while (count--) {
				BMenuItem *item = menu->ItemAt(count);
				BMessage *msg = item->Message();
				BWindow *w = 
					(BWindow*)msg->FindObject(windowsMenuMessageObjName);
				if (w == wind){
					item->SetMarked(check);
					break;
				}
			}
			break;
	}
}
