//-----------------------------------------------------------------------------
//
// TCompile.cpp
//
// This class runs the compiler and linker, in its own blooper threads.
// Multiple threads are used to compile more than one file at once during
// makes. This improves performance by making full use of the multiple CPUs.
//
//-----------------------------------------------------------------------------

#include "TCompile.h"
#include "TWindow.h"
#include "TGlobals.h"
#include "RAlert.h"
#include "RWindow.h"
#include "TApplication.h"
#include "RFile.h"
#include "RDirectory.h"
#include "TProjectInfo.h"

B_DEFINE_CLASS_INFO(TCompile, BLooper);

///TCStatus
//-----------------------------------------------------------------------------

// This class manages the status window.

class TCStatus : public BWindow
{
	B_DECLARE_CLASS_INFO(BWindow);
public:
	TCStatus();
	void ShowMessage(char *statusMsg);
};

B_DEFINE_CLASS_INFO(TCStatus, BWindow);

TCStatus::TCStatus()
	: BWindow(BRect(-1000000, -1000000, -999990, -999983), "", 
		B_BORDERED_WINDOW, B_NOT_RESIZABLE)
{
	Lock();
	BRect viewFrame = Bounds();
	viewFrame.InsetBy(2, 2);
	BStringView *stringView = 
		new BStringView(viewFrame, "stringView", "", B_FOLLOW_ALL);
	AddChild(stringView);
	stringView->SetFontName("Kate");
	Unlock();
}

void TCStatus::ShowMessage(char *statusMsg)
{
	Lock();
	BStringView *stringView = (BStringView*)FindView("stringView");
	stringView->SetText(statusMsg);
	screen_info screenInfo;
	get_screen_info(&screenInfo);
	BRect frame = screenInfo.frame;
	frame.left = frame.right - stringView->StringWidth(statusMsg) - 4;
	frame.bottom = frame.top + 17;
	MoveTo(frame.left, frame.top);
	ResizeTo(frame.right-frame.left, frame.bottom-frame.top);
	if (IsHidden()) Show();
	Activate(FALSE);
	Unlock();
}

///TRunner
//-----------------------------------------------------------------------------

// This blooper class does compiles and links. The TCompile class creates 
// multiple runners so it can do multiple compiles at the same time.

class TRunner : public BLooper
{
	B_DECLARE_CLASS_INFO(BLooper);
public:
	TRunner(long index);
private:
	void MessageReceived(BMessage *msg);
	long index;
};

B_DEFINE_CLASS_INFO(TRunner, BLooper);

TRunner::TRunner(long index)
	: BLooper(B_NORMAL_PRIORITY), index(index)
{
	Run();
}

void TRunner::MessageReceived(BMessage *msg)
{
	extern char **environ;
	
	const char *cmd = msg->FindString("cmd");
	
	RDirectory obj;
	record_ref shref;
	
	project_info.ReadLock();
	project_info.GetObjDirectory(0, &obj);
	get_ref_for_path(project_info.shell, &shref);
	project_info.ReadUnlock();
	
	char objPath[1000];
	obj.GetPath(objPath, 1000);
	
	char tempFileName[B_FILE_NAME_LENGTH+1];
	sprintf(tempFileName, "temp%ld", index);
	
	char str[10000];
	sprintf(str, "%s &>\"%s/%s\"", cmd, objPath, tempFileName);
	
	int argc = 4;
	char **argv = (char**)malloc(20);
	argv[0] = (char*)malloc(3);
	strcpy(argv[0], "sh");
	argv[1] = (char*)malloc(11);
	strcpy(argv[1], "-noprofile");
	argv[2] = (char*)malloc(3);
	strcpy(argv[2], "-c");
	argv[3] = (char*)malloc(strlen(str)+1);
	strcpy(argv[3], str);
	argv[4] = NULL;

	BFile shfile(shref);
		
	thread_id id = load_executable(&shfile, argc, argv, environ);
	long result;
	wait_for_thread(id, &result);
	
	for (char **p = argv; *p; p++) free(*p);
	free(argv);
	
	RFile ofile;
	char *data;
	if (obj.GetFile(tempFileName, &ofile) == B_NO_ERROR) {
		data = (char*)ofile.ReadToMem();
	} else {
		data = "";
	}
	
	BMessage *newMsg = new BMessage(msg->FindLong("what"));
	newMsg->AddLong("index", index);
	newMsg->AddLong("result", result);
	newMsg->AddString("data", data);
	compile_blooper->PostMessage(newMsg);
	
	delete [] data;
}

///TCompile
//-----------------------------------------------------------------------------

TCompile::TCompile()
	: BLooper(B_NORMAL_PRIORITY), status(NULL),
	making(FALSE), compile(NULL), link(NULL), run(FALSE)
{
	for (long i = 0; i < MAX_RUNNERS; i++) {
		TRunnerInfo &info = runnerInfo[i];
		info.runner = NULL;
		info.busy = FALSE;
	}
	status = new TCStatus();
	Run();
}

///Compile
//-----------------------------------------------------------------------------

void TCompile::Compile(record_ref ref)
{
	BMessage *msg = new BMessage(COMPILE_SOURCE_FILE);
	msg->AddRef("ref", ref);
	PostMessage(msg);
}

///MessageReceived
//-----------------------------------------------------------------------------

void TCompile::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case COMPILE_SOURCE_FILE:
			{
				if (!my_app->SaveAll(TRUE)) break;
				if (making) {
					RAlert::InfoMsg("You must wait for the previous make to "
						"finish before trying to do a compile");
					return;
				}
				InitRunners();
				record_ref ref = msg->FindRef("ref");
				RFile file(ref);
				char name[B_FILE_NAME_LENGTH+1];
				file.GetName(name);
				long len = strlen(name);
				name[len-4] = 0;
				long i;
				for (i = 0; i < numRunners; i++) {
					TRunnerInfo &info = runnerInfo[i];
					if (info.busy && strcmp(name, info.name) == 0) {
						RAlert::InfoMsg("That file is already being compiled. "
							"You must wait for the previous compile to finish "
							"before you can start a new one.");
						return;
					}
				}
				if (!StartCompile(ref)) {
					RAlert::InfoMsg("Sorry, all the compile threads are busy. "
						"Try again later.");
					return;
				}
			}
			break;
		case MAKE_PROJECT:
			{
				InitRunners();
				long i;
				for (i = 0; i < numRunners; i++) if (runnerInfo[i].busy) break;
				if (i < numRunners) {
					RAlert::InfoMsg("You must wait for all previous makes and "
						"compiles to finish before you can start a new make.");
					return;
				}
				StartMake(msg);
			}
			break;
		case FINISH_COMPILE:
		case FINISH_LINK:
			{
				long index = msg->FindLong("index");
				TRunnerInfo &info = runnerInfo[index];
				info.busy = FALSE;
				if (info.dead) {
					UpdateStatus();
					break;
				}
				long result = msg->FindLong("result");
				const char *xdata = msg->FindString("data");
				long len = strlen(xdata);
				char *data = new char[len+1];
				strcpy(data, xdata);
				bool ok = msg->what == FINISH_COMPILE ?
					FinishCompile(result, data) :
					FinishLink(result, data);
				delete [] data;
				if (making) {
					if (ok) {
						StartNextMakeStep();
					} else {
						for (long i = 0; i < numRunners; i++) {
							TRunnerInfo &info = runnerInfo[i];
							if (info.busy) info.dead = TRUE;
						}
						making = FALSE;
						UpdateStatus();
					}
				} else {
					UpdateStatus();
				}
			}
			break;
		default:
			inherited::MessageReceived(msg);
			break;
	}
}

///Quit
//-----------------------------------------------------------------------------

void TCompile::Quit()
{
	for (long i = 0; i < numRunners; i++) runnerInfo[i].runner->Quit();
	inherited::Quit();
}


///StartCompile
//-----------------------------------------------------------------------------

bool TCompile::StartCompile(record_ref ref)
{
	long index;
	for (index = 0 ; index < numRunners; index++)
		if (!runnerInfo[index].busy) break;
	if (index >= numRunners) return FALSE;
	
	project_info.ReadLock();

	RFile srcFile(ref);
	char srcName[B_FILE_NAME_LENGTH+1];
	srcFile.GetName(srcName);
	
	char srcPath[1000];
	srcFile.GetPath(srcPath, 1000);
	
	char srcNameNoSuffix[B_FILE_NAME_LENGTH+1];
	strcpy(srcNameNoSuffix, srcName);
	srcNameNoSuffix[strlen(srcName)-4] = 0;
	
	char objPath[1000];
	strcpy(objPath, srcPath);
	char *p = objPath + strlen(objPath) - 1;
	while (p >= objPath && *p != '/') p--;
	*p = 0;
	
	char cmd[2000];
	p = cmd;
	
	sprintf(p, "%s %s", project_info.cc, project_info.beIncludes);
	p += strlen(p);
	
	RDirectory dir;
	long i = 1;
	while (project_info.GetDirectory(i++, &dir) == B_NO_ERROR) {
		char dirPath[1000];
		dir.GetPath(dirPath, 1000);
		sprintf(p, " -I \"%s\"", dirPath);
		p += strlen(p);
	}
	
	sprintf(p, " %s %s -maxerrors 1 -o \"%s/%s/%s.o\" -c \"%s\"", 
		project_info.includes, project_info.cflags, objPath, 
		project_info.obj, srcNameNoSuffix, srcPath);
	p += strlen(p);
	
	BMessage *msg = new BMessage('doit');
	msg->AddString("cmd", cmd);
	msg->AddLong("what", FINISH_COMPILE);
	
	TRunnerInfo &info = runnerInfo[index];
	info.runner->PostMessage(msg);
	info.busy = TRUE;
	info.dead = FALSE;
	strcpy(info.name, srcNameNoSuffix);
	info.linking = FALSE;
	
	UpdateStatus();
	
	project_info.ReadUnlock();
	
	return TRUE;
}

///FinishCompile
//-----------------------------------------------------------------------------

bool TCompile::FinishCompile(long result, char *data)
{
	long size = strlen(data);
		
	bool compilerError = result != 0 || size > 0;
	bool messageSentToApp = FALSE;
		
	if (compilerError && (result == 0 || result == 256)) {
		while (TRUE) {
			char *p = data;
			while (*p && strncmp(p, "(line", 5) != 0) p++;
			if (!*p) break;
			char x = *(p-1);
			*(p-1) = 0;
			char *refBegin = p-2;
			while (refBegin >= data && *refBegin != B_RETURN) refBegin--;
			if (refBegin >= data) refBegin++;
			record_ref ref;
			long err = get_ref_for_path(refBegin, &ref);
			*(p-1) = x;
			if (err) break;
			long line = atol(p+6);
			if (line == 0) break;
			char *s = p;
			while (TRUE) {
				while (*s && *s != B_RETURN) s++;
				if (!*s) break;
				s++;
				if (strncmp(s, "###", 3) == 0) break;
			}
			if (*s || *(s-1) == B_RETURN) s--;
			char *r = s-1;
			while (r > p && *r != B_RETURN) r--;
			if (r <= p) break;
			char *q = r-1;
			while (q > p && *q != B_RETURN) q--;
			if (q <= p) break;
			for (char *z = p; z < q; z++) if (*z == B_RETURN) *z = ' ';
			*q = *r = *s = 0;
			BMessage *msg = new BMessage(B_REFS_RECEIVED);
			msg->AddRef("refs", ref);
			msg->AddLong("line", line);
			msg->AddString("emess1", p);
			msg->AddString("emess2", q+1);
			msg->AddString("emess3", r+1);
			be_app->PostMessage(msg);
			messageSentToApp = TRUE;
			break;
		}
	}
	
	if (compilerError && !messageSentToApp) {
		
		char *text = new char[100000];
		char *p = text;
		sprintf(p, "Compiler result code = %ld\n\nCompiler output = \n\n",
			result);
		p += strlen(p);
		strcpy(p, data);
		p += strlen(p);
		
		new TWindow(text, p-text, FALSE, -1, -1, "Unexpected Compiler Output");
		
		delete [] text;
	}
	
	return !messageSentToApp;
}

///StartLink
//-----------------------------------------------------------------------------

void TCompile::StartLink()
{
	char cmd[10000];
	char *p = cmd;
	
	project_info.ReadLock();
	
	RDirectory primary;
	project_info.GetDirectory(0, &primary);
	char primaryPath[1000];
	primary.GetPath(primaryPath, 1000);
	
	sprintf(p, "%s -o \"%s/%s\"", project_info.ld, primaryPath, 
		project_info.app);
	p += strlen(p);
	
	long i = 0;
	while (char *name = (char*)link->ItemAt(i++)) {
		RFile file;
		if (project_info.GetObjFile(name, &file) == B_NO_ERROR) {
			char objPath[1000];
			file.GetPath(objPath, 1000);
			sprintf(p, " \"%s\"", objPath);
			p += strlen(p);
		}
		delete [] name;
	}
	
	sprintf(p, " %s", project_info.ldflags);
	p += strlen(p);
	
	link->MakeEmpty();
	
	BMessage *msg = new BMessage('doit');
	msg->AddString("cmd", cmd);
	msg->AddLong("what", FINISH_LINK);
	
	TRunnerInfo &info = runnerInfo[0];
	info.runner->PostMessage(msg);
	info.busy = TRUE;
	info.dead = FALSE;
	strcpy(info.name, project_info.app);
	info.linking = TRUE;
	
	project_info.ReadUnlock();
	
	UpdateStatus();
}

///FinishLink
//-----------------------------------------------------------------------------

bool TCompile::FinishLink(long result, char *data)
{
	long size = strlen(data);
		
	bool linkErr = result != 0 || size > 0;
	
	if (result == 0 && linkErr) {
		if (strcmp(data, "## link warning : ignored: "
			"'__sinit' (descriptor) in start_dyn.c, "
			"previously defined in libbe.so\n") == 0) 
				linkErr = FALSE;
	}
	
	if (linkErr) {
	
		char *text = new char[100000];
		char *p = text;
		sprintf(p, "Linker result code = %ld\n\n\Linker output = \n\n",
			result);
		p += strlen(p);
		strcpy(p, data);
		p += strlen(p);
		
		new TWindow(text, p-text, FALSE, -1, -1, "Linker Error");
		
		delete [] text;
	}
	
	return !linkErr;
}

///RunApp
//-----------------------------------------------------------------------------

void TCompile::RunApp()
{
	extern char **environ;
	
	project_info.ReadLock();

	RFile app;
	if (project_info.GetPrimaryFile(project_info.app, &app) == B_NO_ERROR) {
		int argc = 1;
		char **argv = (char**)malloc(8);
		argv[0] = (char*)malloc(strlen(project_info.app) + 1);
		strcpy(argv[0], project_info.app);
		argv[1] = NULL;
		thread_id id = load_executable(&app, argc, argv, environ);
		resume_thread(id);
	}
	
	project_info.ReadUnlock();
}

///StartMake
//-----------------------------------------------------------------------------

void TCompile::StartMake(BMessage *msg)
{
	making = TRUE;
	
	if (compile) {
		long i = 0;
		while (record_ref *ref = (record_ref*)compile->ItemAt(i++))
			delete ref;
		delete compile;
	}
	
	if (link) {
		long i = 0;
		while (char *name = (char*)link->ItemAt(i++))
			delete [] name;
		delete link;
	}
	
	compile = (BList*)msg->FindObject("compile");
	link = (BList*)msg->FindObject("link");
	run = msg->FindBool("run");
	
	StartNextMakeStep();
}

///StartNextMakeStep
//-----------------------------------------------------------------------------

void TCompile::StartNextMakeStep()
{
	if (compile->CountItems() > 0) {
		while (record_ref *ref = (record_ref*)compile->FirstItem()) {
			if (!StartCompile(*ref)) break;
			delete ref;
			long i = 0;
			compile->RemoveItem(i);
		}
	} else if (link->CountItems() > 0) {
		long i;
		for (i = 0; i < numRunners; i++) {
			if (runnerInfo[i].busy) break;
		}
		if (i >= numRunners) {
			StartLink();
		} else {
			UpdateStatus();
		}
	} else {
		if (run) RunApp();
		making = FALSE;
		UpdateStatus();
	}
}

///InitRunners
//-----------------------------------------------------------------------------

void TCompile::InitRunners()
{
	project_info.ReadLock();
	numRunners = project_info.numCompileThreads;
	project_info.ReadUnlock();
	for (long i = 0; i < numRunners; i++) {
		TRunnerInfo &info = runnerInfo[i];
		if (!info.runner) info.runner = new TRunner(i);
	}
}

///UpdateStatus
//-----------------------------------------------------------------------------

void TCompile::UpdateStatus()
{
	long numBusy = 0;
	char statusMsg[2000];
	char *p = statusMsg;
	for (long i = 0; i < numRunners; i++) {
		TRunnerInfo &info = runnerInfo[i];
		if (info.busy) {
			if (numBusy == 0) {
				sprintf(p, "%s %s", info.linking ? "Linking" : "Compiling",
					info.name);
			} else {
				sprintf(p, "/%s", info.name);
			}
			p += strlen(p);
			numBusy++;
		}
	}
	if (numBusy > 0) {
		status->ShowMessage(statusMsg);
	} else {
		if (!status->IsHidden()) status->Hide();
	}
}
