//-----------------------------------------------------------------------------
//
// TDepend.cpp
//
// Header file dependencies class and blooper thread.
//
// All of the data structures used to track dependencies are read and written
// only by code running in the TDepend thread. So no lock is needed.
//
//-----------------------------------------------------------------------------

#include "TDepend.h"
#include "TGlobals.h"
#include "TWindow.h"
#include "RAlert.h"
#include "TUtil.h"
#include "TCompile.h"
#include "RFile.h"
#include "RDirectory.h"
#include "TProjectInfo.h"
#include "RUtil.h"

B_DEFINE_CLASS_INFO(TDepend, BLooper);

///TIncludes
//-----------------------------------------------------------------------------

// This class records the cached header file include info from the database.
//
// The database is used to record header file include information. The table
// name is "BetterEdit Include Info Cache". There are two fields in the table:
//
// Field "dirid" is a recordID field. It contains the record id of the project
// directory (primary or secondary). This field is indexed.
//
// Field "includes" is a string field. This field records all the cached
// include info for each .cpp and .h file in the directory, in the following 
// format:
//
// 		file name
//			timeStamp
//			included .h file name
//			... repeat as needed for more included .h files
//		... repeat as needed for more .cpp and .h files
//
// This use of the database is weak and wimpy. It does not use very good
// database technology. This is a remnant of the use of ".includes" files
// in the "obj" directories in beta versions of BetterEdit to record this
// information. I was too lazy to do it "the right way" when I switched over
// to using the database.
//
// On the other hand, the current code works just fine and is quite efficient.
// It's just ugly, that's all.

class TIncludes : public BObject
{
	B_DECLARE_CLASS_INFO(BObject);
public:
	TIncludes(const char *name, const long dirIndex, const long timeStamp);
	
	const char *name;	// file name - points into includesData
	long dirIndex;		// index of project dir to which this entry corresponds
	long timeStamp;		// time stamp
	BList inc;			// direct includes, type list of char* - points into
						// includesData
	bool defunct;		// TRUE if defunct: TDepend::modules list has current 
						// correct info.
};

B_DEFINE_CLASS_INFO(TIncludes, BObject);

TIncludes::TIncludes(const char *name, const long dirIndex, 
	const long timeStamp)
	: name(name), dirIndex(dirIndex), timeStamp(timeStamp), inc(), 
	defunct(FALSE)
{
}	

///TModule
//-----------------------------------------------------------------------------

// This class keeps track of information for each module: the app file,
// each .o file, and all the .cpp and .h files.

typedef enum {
	appfile,
	doto,
	dotcpp,
	doth
} TModuleKind;

class TModule : public BObject 
{
	B_DECLARE_CLASS_INFO(BObject);
public:
	TModule(const TModuleKind kind, const char *fileName, const long dirIndex);
	~TModule();
	
	TModuleKind kind;	// kind of module
	char *name;			// file name
	long dirIndex;		// index of project dir from which this entry came
	long mod;			// last mod time during dirty calc
	BList dep;			// direct dependencies (app and .o only),
						// type list of TModule*
	BList inc;			// direct includes (.cpp and .h only),
						// type list of TModule*
	TModule *partner;	// .cpp <--> .o partner pointer
	bool dirty;			// true if dirty (app and .o only)
	bool seen;			// true if seen during dirty calc
	long timeStamp;		// time stamp for .cpp and .h files, 0 if unknown
};

B_DEFINE_CLASS_INFO(TModule, BObject);

TModule::TModule(const TModuleKind kind, const char *fileName,
	const long dirIndex)
	: kind(kind), dirIndex(dirIndex), mod(0), dep(), inc(), partner(NULL),
	dirty(FALSE), seen(FALSE), timeStamp(0)
{
	name = new char[strlen(fileName)+1];
	strcpy(name, fileName);
}

TModule::~TModule()
{
	delete [] name;
}

///TDepend
//-----------------------------------------------------------------------------

TDepend::TDepend()
	: BLooper(B_NORMAL_PRIORITY), modules(), dependsInited(FALSE),
	includes(), haveIncludes(FALSE)
{
	Run();
}

///SetProject
//-----------------------------------------------------------------------------

void TDepend::SetProject(record_ref ref)
{
	// Sets the project directory. 
	//
	// Called by the "Set Project" command, when the first .cpp or .h
	// file is opened or saved, and whenever a modified "project" file
	// is saved.
	//
	// Sets the project directory, reads the "project" files, and
	// initialized TDepend's internal data structures.

	BMessage *msg = new BMessage('sprj');
	msg->AddRef("ref", ref);
	PostMessage(msg);
}

void TDepend::SetProject()
{
	// Rereads the "/system/project" file. Called if a "project" file
	// is saved but we don't have a project directory yet.
	
	record_ref ref = {-1,-1};
	SetProject(ref);
}

///CreateMakeFile
//-----------------------------------------------------------------------------

void TDepend::CreateMakeFile()
{
	PostMessage(new BMessage('cmkf'));
}

///UpdateIncludeFileDependencies
//-----------------------------------------------------------------------------

void TDepend::UpdateIncludeFileDependencies(const record_ref ref,
	const char *text, const long len)
{
	// Called when a .cpp or .h project file is saved.
	//
	// The record of include file dependencies is updated in the module
	// list.
	//
	// Note that the text is scanned for #include directives in the
	// caller's thread, and then a message is posted to the TDepend thread
	// to update the module list.

	BMessage *msg = BuildMessageListingIncludes(text, len);
	msg->AddRef("ref", ref);
	PostMessage(msg);
}

///Make
//-----------------------------------------------------------------------------

void TDepend::Make()
{
	PostMessage(new BMessage('make'));
}

///MakeAndRun
//-----------------------------------------------------------------------------

void TDepend::MakeAndRun()
{
	BMessage *msg = new BMessage('make');
	msg->AddBool("run", TRUE);
	PostMessage(msg);
}

///WriteIncludesAndQuit
//-----------------------------------------------------------------------------

void TDepend::WriteIncludesAndQuit()
{
	PostMessage(new BMessage('wrdi'));
}

///MessageReceived
//-----------------------------------------------------------------------------

void TDepend::MessageReceived(BMessage *msg)
{
	switch (msg->what) {	
		case 'sprj':		
			DoSetProject(msg->FindRef("ref"));
			break;			
		case 'cmkf':
			project_info.ReadLock();
			DoCreateMakeFile();
			project_info.ReadUnlock();
			break;			
		case 'upfd':
			project_info.ReadLock();
			DoUpdateIncludeFileDependencies(msg);
			project_info.ReadUnlock();
			break;			
		case 'make':
			project_info.ReadLock();
			DoMake(msg->FindBool("run"));
			project_info.ReadUnlock();
			break;
		case 'wrdi':
			project_info.ReadLock();
			DoWriteIncludes();
			project_info.ReadUnlock();
			Quit();
			break;
		default:
			inherited::MessageReceived(msg);
	}
}

///DoSetProject
//-----------------------------------------------------------------------------

void TDepend::DoSetProject(record_ref ref)
{
	// We construct the new project info in the "info" local variable. Only
	// when it has been completely constructed and all error checking is
	// complete do we copy it to the global project_info. If any kind of 
	// error occurs, the old project info is unchanged.
	//
	// This function uses the TProjectInfo destructor and assignment
	// operator. They're important.

	TProjectInfo info;

	if (ref.database >= 0) 
		if (!info.SetProjectInfo(ref)) return;
	
	project_info.WriteLock();
	project_info = info;
	project_info.WriteUnlock();
	
	if (ref.database < 0) return;
	
	// Nuke the old project dependency information.
	
	long i = 0;
	while (TModule *m = (TModule*)modules.ItemAt(i++)) delete m;
	modules.MakeEmpty();
	dependsInited = FALSE;
	i = 0;
	while (TIncludes *z = (TIncludes*)includes.ItemAt(i++))
		delete z;
	includes.MakeEmpty();
	i = 0;
	while (char *data = (char*)includesData.ItemAt(i++)) delete [] data;
	includesData.MakeEmpty();
	haveIncludes = FALSE;
	
	// Make the new app module the first module in the module list.
		
	project_info.ReadLock();
	TModule *m = new TModule(appfile, project_info.app, -1);
	project_info.ReadUnlock();
	modules.AddItem(m);
	
	// Add open commands to the main menu for each project directory.
	// Send a message to the app thread and let it do this job.
	
	be_app->PostMessage(new BMessage(ADD_PROJECT_OPEN_COMMANDS));
}

///DoMake
//-----------------------------------------------------------------------------

void TDepend::DoMake(bool run)
{
	InitDependencies();

	// Get the mod date for each module
	
	long i = 0;
	while (TModule *m = (TModule*)modules.ItemAt(i++)) {
		m->mod = 0;
		RFile file;
		long err;
		if (m->kind == doto) {
			err = project_info.GetObjFile(m->name, &file);
		} else {
			err = project_info.GetFile(m->name, &file);
		}
		if (err == B_NO_ERROR) m->mod = file.ModificationTime();
	}
	
	// Compute the dirty flags. 
	//
	// Definitions:
	//
	// x "depends on" y if x depends on y directly or indirectly.
	//
	// x "incudes" y if x includes y directly or indirectly.
	//
	// x is "dirty" if :
	// any y on which x depends is newer than x or does not exist (mod date 
	//		is zero),
	// or any z which any y includes where x depends on y is newer than x
	// 		or does not exist.
	
	i = 0;
	while (TModule *m = (TModule*)modules.ItemAt(i++)) {
		long j = 0;
		while (TModule *x = (TModule*)modules.ItemAt(j++))
			x->seen = FALSE;
		m->dirty = ModDateNewer(m->mod, m->dep);
	}
			
	// If this is not a "Run" command, and if the app is not dirty, then issue 
	// an alert.
	
	TModule *app = (TModule*)modules.FirstItem();
	if (!run && !app->dirty) {
		RAlert::InfoMsg("The project is already up-to-date!\n ");
		return;
	}
	
	// Build a message and send it to the compile blooper to do do the
	// compile/link/run. 
	//
	// The record refs for the .cpp files which need to to be compiled are 
	// listed in a BList object which is sent as the "compile" parameter.
	// This list may be empty.
	//
	// The names for the .o files which need to be linked are listed in a
	// BList object which is sent as the "link" parameter. Names are used
	// instead of record refs because the .o files may not exist yet. This
	// list may be empty.
	//
	// The "run" parameter is a bool which tells the compile blooper whether
	// the program should be loaded and executed after it is made. 
	
	BMessage *msg = new BMessage(MAKE_PROJECT);
	
	BList *compile = new BList();
	BList *link = new BList();
	
	i = 0;
	while (TModule *m = (TModule*)modules.ItemAt(i++)) {
		if (m->dirty) {
			switch (m->kind) {
				case appfile:
					BList &dep = app->dep;
					long j = 0;
					while (TModule *d = (TModule*)dep.ItemAt(j++)) {
						long len = strlen(d->name);
						char *name = new char[len+1];
						strcpy(name, d->name);
						link->AddItem(name);
					}
					break;
				case doto:
					record_ref *ref = new record_ref;
					RFile file;
					if (project_info.GetFile(m->partner->name, &file) == 
						B_NO_ERROR) 
					{
						*ref = file.Ref();
						compile->AddItem(ref);
					}
					break;
			}
		}
	}
	
	msg->AddObject("compile", compile);
	msg->AddObject("link", link);
	msg->AddBool("run", run);
	
	compile_blooper->PostMessage(msg);
}

///   ModDateNewer
//-----------------------------------------------------------------------------

bool TDepend::ModDateNewer(long mod, BList &dep)
{
	long i = 0;
	while (TModule *d = (TModule*)dep.ItemAt(i++)) {
		if (!d->seen) {
			d->seen = TRUE;
			if (mod <= d->mod || (d->mod == 0 && d->kind == doto)) return TRUE;
			if (ModDateNewer(mod, d->dep)) return TRUE;
			if (ModDateNewer(mod, d->inc)) return TRUE;
		}
	}
	return FALSE;
}

///DoCreateMakeFile
//-----------------------------------------------------------------------------

void TDepend::DoCreateMakeFile()
{	
	InitDependencies();
	
	char *data = new char[100000];
	char *p = data;
	
	sprintf(p,
"#\n# \tmakefile for %s\n#\n\nSHELL\t\t= %s\nINCLUDES \t= %s\nOBJ\t\t\t= %s\nAPP\t\t\t= %s\n\nCC\t\t\t= %s\nCFLAGS\t\t= %s\nLD \t\t\t= %s\nLDFLAGS\t\t= %s\n\nOBJS \t\t= $(CPPOBJS)\n\n",
		project_info.app, project_info.shell, project_info.includes, 
		project_info.obj, project_info.app, project_info.cc, 
		project_info.cflags, project_info.ld, project_info.ldflags);
	p += strlen(p);
	
	long i = 1;
	long numSecondaries = 0;
	RDirectory dir;
	while (project_info.GetDirectory(i++, &dir) == B_NO_ERROR) {
		char dirPath[1000];
		dir.GetPath(dirPath, 1000);
		char *q = dirPath;
		while (*q && *q != '/') q++;
		sprintf(p, "DIR%ld\t\t= %s\n", i-1, q);
		p += strlen(p);
		numSecondaries++;
	}
	if (numSecondaries > 0) {
		sprintf(p, "XINCLUDES\t=");
		p += strlen(p);
		for (i = 1; i <= numSecondaries; i++) {
			sprintf(p, " -I $(DIR%ld)", i);
			p += strlen(p);
		}
		*p++ = B_RETURN;
		*p++ = B_RETURN;
	}

	p = WriteCPPOBJS(p);
	
	strcpy(p,
"\nall:\t$(OBJ) $(APP)\n\t@echo Just made $(notdir $(APP))\n\nclean:\n\t@echo Removing $(OBJS)\n\t@-rm -f $(OBJS)\n\n$(APP): $(OBJ) $(OBJS)\n\t$(LD) -o $(APP) $(OBJS) $(LDFLAGS)\n\n$(OBJ):\n\t-mkdir $(OBJ)\n\n"
	);
	p += strlen(p);
	
	p = WriteOBJS(p, numSecondaries > 0);
	
	new TWindow(data, p-data, FALSE, -1, -1, "makefile");
	
	delete [] data;
}

///   WriteCPPOBJS
//-----------------------------------------------------------------------------

char *TDepend::WriteCPPOBJS(char *p)
{
	strcpy(p, "CPPOBJS\t\t=");
	p += strlen(p);
	
	TModule *app = (TModule*)modules.FirstItem();
	BList &dep = app->dep;
	long i = 0;
	while (TModule *d = (TModule*)dep.ItemAt(i++)) {
		if (d->partner->dirIndex > 0) {
			sprintf(p, "\t\\\n\t$(DIR%ld)/$(OBJ)/%s", d->partner->dirIndex,
				d->name);
		} else {
			sprintf(p, "\t\\\n\t$(OBJ)/%s", d->name);
		}
		p += strlen(p);
	}
	*p++ = B_RETURN;
	
	return p;
}

///   WriteOBJS
//-----------------------------------------------------------------------------

char *TDepend::WriteOBJS(char *p, bool xincludes)
{
	long i = 0;
	while (TModule *m = (TModule*)modules.ItemAt(i++)) {
		if (m->kind == doto) {
			long dirIndex = m->partner->dirIndex;
			if (dirIndex > 0) {
				sprintf(p, "$(DIR%ld)/$(OBJ)/%s:", dirIndex, m->name);
			} else {
				sprintf(p, "$(OBJ)/%s:", m->name);
			}
			p += strlen(p);	
			long j = 0;
			while (TModule *x = (TModule*)modules.ItemAt(j++))
				x->seen = FALSE;
			p = WriteDeps(m->dep, p);
			sprintf(p, 
				"\n\t$(CC) %s $(INCLUDES) $(CFLAGS) -o $@ -c ",
				xincludes ? "$(XINCLUDES)" : "");
			p += strlen(p);
			if (dirIndex > 0) {
				sprintf(p, "$(DIR%ld)/%s\n\n", dirIndex, m->partner->name);
			} else {
				sprintf(p, "%s\n\n", m->partner->name);
			}
			p += strlen(p);
		}
	}
	
	return p;
}

///   WriteDeps
//-----------------------------------------------------------------------------

char *TDepend::WriteDeps(BList &dep, char *p)
{
	long i = 0;
	while (TModule *d = (TModule*)dep.ItemAt(i++)) {
		if (!d->seen) {
			if (d->dirIndex > 0) {
				sprintf(p, " $(DIR%ld)/%s", d->dirIndex, d->name);
			} else {
				sprintf(p, " %s", d->name);
			}
			p += strlen(p);
			d->seen = TRUE;
			p = WriteDeps(d->inc, p);
		}
	}
	return p;
}

///InitDependencies
//-----------------------------------------------------------------------------

void TDepend::InitDependencies()
{
	// Initialize the full module list and make certain all dependencies
	// are up-to-date. This is needed for the "Make" and "Create Makefile"
	// commands. There's no need to do this if the user never uses either
	// of these commands, so we postpone doing it until one of these commands
	// is used, if ever. Once we do it, we never have to do it again.

	if (dependsInited) return;
	dependsInited = TRUE;
	
	// Scan the project directory and add any .cpp/.h/.o modules
	// which haven't already been added.
	
	RFile file;
	long i = 0;
	long dirIndex;
	while (project_info.GetFile(i++, &file, &dirIndex) == B_NO_ERROR) {
		char name[B_FILE_NAME_LENGTH+1];
		file.GetName(name);
		if (IsDotCpp(name) || IsDotH(name)) {
			TModule *m = FindModule(name);
			if (!m || m->timeStamp == 0) {
			
				BMessage *msg;
				
				// First try to get valid include info from the
				// includes cache.
			
				ReadIncludes();
				long j = 0;
				TIncludes *z;
				while (z = (TIncludes*)includes.ItemAt(j++)) {
					if (z->dirIndex == dirIndex && strcmp(name, z->name) == 0) {
						if (file.ModificationTime() <= z->timeStamp) {
							msg = new BMessage('upfd');
							BList &inc = z->inc;
							long k = 0;
							while (char *q = (char*)inc.ItemAt(k++))
								msg->AddString("include", q);
						} else {
							z = NULL;
						}
						break;
					}
				}
				
				// If we don't have any include info on the includes
				// cache, or if it's stale, we have to read the file
				// and scan it for #include directives. This should only 
				// happen if the database was rebuilt or is damaged, or if a 
				// different editor was used to modify the file, or if a new 
				// file was dragged into the project folder from some other 
				// source.
				
				if (!z) {
					long len;
					char *text = (char*)file.ReadToMem(&len);
					msg = BuildMessageListingIncludes(text, len);
					delete [] text;
				}
				
				// Update the include file dependencies for this module.
					
				msg->AddRef("ref", file.Ref());				
				DoUpdateIncludeFileDependencies(msg);
				delete msg;
			}
		}
	}
}

///DoUpdateIncludeFileDependencies
//-----------------------------------------------------------------------------

void TDepend::DoUpdateIncludeFileDependencies(BMessage *msg)
{
	// Update the include file dependencies for a .cpp or .h module.
	// The "ref" parameter in "msg" specifies the module to be updated.
	// The "include" parameters list the names of the .h files included by
	// this file.

	record_ref ref = msg->FindRef("ref");
	RFile file(ref);
	char name[B_FILE_NAME_LENGTH+1];
	file.GetName(name);
	TModule *m = FindModule(name);
	
	// Add the module to the modules list if it's not there already.
	
	if (!m) {
		long dirIndex;
		if (project_info.GetFile(name, &file, &dirIndex) == B_NO_ERROR) {
			m = new TModule(IsDotCpp(name) ? dotcpp : doth, name, dirIndex);
			modules.AddItem(m);
			if (m->kind == dotcpp) {
				long len = strlen(name);
				strcpy(name+len-4, ".o");
				TModule *mo = new TModule(doto, name, -1);
				m->partner = mo;
				mo->partner = m;
				mo->dep.AddItem(m);
				modules.AddItem(mo);
				TModule *app = (TModule*)modules.FirstItem();
				app->dep.AddItem(mo);
			}
		}
	}
	
	// Update the includes list for this module.
	
	long i = 0;
	BList &inc = m->inc;
	inc.MakeEmpty();
	while (const char *name = msg->FindString("include", i++)) {
		TModule *x = FindModule(name);
		if (!x) {
			RFile file;
			long dirIndex;
			if (project_info.GetFile(name, &file, &dirIndex) == B_NO_ERROR) {
				x = new TModule(doth, name, dirIndex);
			} else {
				x = new TModule(doth, name, -1);
			}
			modules.AddItem(x);
		}
		inc.AddItem(x);
	}
	m->timeStamp = file.ModificationTime();
}

///BuildMessageListingIncludes
//-----------------------------------------------------------------------------

BMessage *TDepend::BuildMessageListingIncludes(const char *text, 
	const long len)
{	
	// Builds and returns a BMessage listing the names of all the .h
	// files included in the test.

	const char *prefix = "#include \"";
	const prefixLen = 10;
	
	BMessage *msg = new BMessage('upfd');

	const char *pEnd = text+len;
	for (const char *p = text; p < pEnd-prefixLen; p++) {
		if (strncmp(p, prefix, prefixLen) == 0) {
			p += prefixLen;
			char name[B_FILE_NAME_LENGTH+1];
			char *q = name;
			while (p < pEnd && *p != '"' && *p != B_RETURN && 
				q < name+B_FILE_NAME_LENGTH) *q++ = *p++;
			*q = 0;
			if (IsDotH(name)) msg->AddString("include", name);
		}
		while (*p && *p != B_RETURN) p++;
		if (!*p) break;
	}
	
	return msg;
}

///FindModule
//-----------------------------------------------------------------------------

TModule *TDepend::FindModule(const char *name)
{
	long i = 0;
	TModule *m;
	
	while (m = (TModule*)modules.ItemAt(i++)) {
		if (strcmp(name, m->name) == 0) break;
	}
	return m;
}

///ReadIncludes
//-----------------------------------------------------------------------------

void TDepend::ReadIncludes()
{
	// Reads the cached header file include info from the database into
	// the "includes" and "includesData" internal data structures.

	if (haveIncludes) return;
	haveIncludes = TRUE;
	long i = 0;

	RDirectory dir;
	while (project_info.GetDirectory(i++, &dir) == B_NO_ERROR) {
		record_ref dirRef = dir.Ref();
		BDatabase *d = database_for(dirRef.database);
		if (!d) continue;
		d->Sync();
		BTable *t = d->FindTable("BetterEdit Include Info Cache");
		if (!t) continue;
		BQuery *q = new BQuery();
		q->AddTable(t);
		q->PushField("dirid");
		q->PushArg(dirRef.record);
		q->PushOp(B_EQ);
		q->FetchOne();
		BRecord *r;
		if (q->CountRecordIDs()) {
			BRecord *r;
			r = new BRecord(q->Database(), q->RecordIDAt(0));
			const char *d = r->FindString("includes");
			if (!r->Error()) {
				char *data = new char[strlen(d)+1];
				strcpy(data, d);
				includesData.AddItem(data);
				char *p = data;
				while (*p) {
					char *q;
					q = p;
					while (*q && *q != B_RETURN) q++;
					*q = 0;
					if (p == q || q - p > B_FILE_NAME_LENGTH) break;
					q++;
					while (*q == B_TAB || *q == ' ') q++;
					long timeStamp = atol(q);
					if (timeStamp == 0) break;
					while (*q && *q != B_RETURN) q++;
					TIncludes *z = new TIncludes(p, i-1, timeStamp);
					includes.AddItem(z);
					BList &inc = z->inc;
					p = q + 1;
					while (*p == B_TAB || *p == ' ') {
						while (*p == B_TAB || *p == ' ') p++;
						q = p;
						while (*q && *q != B_RETURN) q++;
						*q = 0;
						if (p == q || q - p > B_FILE_NAME_LENGTH) break;
						inc.AddItem(p);
						p = q+1;
					}
				}
			}
			delete r;
		}
		delete q;
	}
}

///DoWriteIncludes
//-----------------------------------------------------------------------------

void TDepend::DoWriteIncludes()
{
	// Called at program termination to update the database includes info 
	// cache.
	
	if (!project_info.haveProject) return;
	
	if (modules.CountItems() == 0) return;
	
	ReadIncludes();
	
	long i = 0;
	RDirectory dir;
	while (project_info.GetDirectory(i++, &dir) == B_NO_ERROR) {
	
		// Create a big buffer to hold the new data.
		
		char *newIncludes = new char[100000];
		char *p = newIncludes;
		
		// For each .cpp and .h file in our module list which we have
		// have scanned and is from this directory, build the new entry for 
		// the file and add it to the buffer.
		
		long j = 0;
		while (TModule *m = (TModule*)modules.ItemAt(j++)) {
		
			if ((m->kind == dotcpp || m->kind == doth) && 
				m->dirIndex == i-1 && m->timeStamp > 0) 
			{
				RFile file;
				if (project_info.GetFile(m->name, &file) == B_NO_ERROR) {
				
					// Print the new entry to the buffer.
				
					sprintf(p, "%s\n", m->name);
					p += strlen(p);
					sprintf(p, "\t%ld\n", m->timeStamp);
					p += strlen(p);
					long j = 0;
					BList &inc = m->inc;
					while (TModule *x = (TModule*)inc.ItemAt(j++)) {
						sprintf(p, "\t%s\n", x->name);
						p += strlen(p);
					}
					
					// Mark any corresponding entry in the old includes
					// list as defunct.
					
					long k = 0;
					TIncludes *z;
					while (z = (TIncludes*)includes.ItemAt(k++)) {
						if (z->dirIndex == m->dirIndex &&
							strcmp(m->name, z->name) == 0) 
						{
							z->defunct = TRUE;
							break;
						}
					}
				}
			}
		}
		
		// Copy unmarked entries for this directory from the includes 
		// list to the new buffer, provided the file still exists and hasn't 			
		// been modified since the last time we saw it.
		
		j = 0;
		while (TIncludes *z = (TIncludes*)includes.ItemAt(j++)) {
			RFile file;
			if (z->dirIndex == i-1 && !z->defunct &&
				project_info.GetFile(z->name, &file) == B_NO_ERROR &&
				file.ModificationTime() <= z->timeStamp) 
			{
				sprintf(p, "%s\n", z->name);
				p += strlen(p);
				sprintf(p, "\t%ld\n", z->timeStamp);
				p += strlen(p);
				long k = 0;
				BList &inc = z->inc;
				while (char *q = (char*)inc.ItemAt(k++)) {
					sprintf(p, "\t%s\n", q);
					p += strlen(p);
				}
			}
		}

		// Write the new database record for this directory.
		
		*p++ = 0;
		record_ref dirRef = dir.Ref();
		BDatabase *d = database_for(dirRef.database);
		if (d) {
			d->Sync();
			BTable *t = d->FindTable("BetterEdit Include Info Cache");
			if (!t) {
				t = d->CreateTable("BetterEdit Include Info Cache");
				if (t) {
					t->AddRecordIDField("dirid", B_INDEXED_FIELD);
					t->AddStringField("includes");
				}
			}
			if (t) {
				BQuery *q = new BQuery();
				q->AddTable(t);
				q->PushField("dirid");
				q->PushArg(dirRef.record);
				q->PushOp(B_EQ);
				q->FetchOne();
				BRecord *r;
				if (q->CountRecordIDs()) {
					r = new BRecord(q->Database(), q->RecordIDAt(0));
				} else {
					r = new BRecord(t);
					r->SetRecordID("dirid", dirRef.record);
				}
				r->SetString("includes", newIncludes);
				r->Commit();
				delete r;
				delete q;
			}
		}
		delete [] newIncludes;
		
	}
}
