//-----------------------------------------------------------------------------
//
// TFindInfo.cpp
//
// "Find" information class.
//
// Shared by threads. Must be locked whenever accessed via the "Lock" and 
// "Unlock" member functions. The app thread and window threads may obtain the // lock. To prevent deadlocks, while you have the lock, you must not lock the 
// app or any windows.
//
//-----------------------------------------------------------------------------

#include "TFindInfo.h"
#include "TGlobals.h"
#include "RFile.h"
#include "RDirectory.h"
#include "TProjectInfo.h"

B_DEFINE_CLASS_INFO(TFindInfo, BObject);

///TFindInfo
//-----------------------------------------------------------------------------

TFindInfo::TFindInfo()
	:caseSensitive(FALSE),
	wholeWord(FALSE),
	wrap(FALSE),
	forward(TRUE),
	scope(FIND_THIS_FILE),
	pos(-1)
{
	searchString = new char[1];
	*searchString = 0;
	replaceString = new char[1];
	*replaceString = 0;
}

///Lock
//-----------------------------------------------------------------------------

void TFindInfo::Lock()
{
	lock.Lock();
}

///Unlock
//-----------------------------------------------------------------------------

void TFindInfo::Unlock()
{
	lock.Unlock();
}

///BuildFileList
//-----------------------------------------------------------------------------

void TFindInfo::BuildFileList(record_ref parentRef)
{
	// Caller must have lock.

	pos = -1;
	long i = 0;
	while (TFindFileListElement *el = (TFindFileListElement*)list.ItemAt(i++))
		delete el;
	list.MakeEmpty();
	if (scope == FIND_STD_C_HEADER_FILES) {
		char incPath[1000];
		project_info.ReadLock();
		sprintf(incPath, "%s%s", project_info.beHeaders, "/inc");
		project_info.ReadUnlock();
		record_ref ref;
		if (get_ref_for_path(incPath, &ref) == B_NO_ERROR) {
			RDirectory dir(ref);
			if (!dir.Error()) AddDirContentsToFileList(&dir);
		}
	} else if (scope == FIND_BE_HEADER_FILES) {
		record_ref ref;
		project_info.ReadLock();
		long err = get_ref_for_path(project_info.beHeaders, &ref);
		project_info.ReadUnlock();
		if (!err) {
			RDirectory bedir(ref);
			if (!bedir.Error()) {
				long i = 0;
				RDirectory dir;
				while (bedir.GetDirectory(i++, &dir) == B_NO_ERROR) {
					char dirName[B_FILE_NAME_LENGTH+1];
					dir.GetName(dirName);
					if (strcmp(dirName, "inc") != 0)
						AddDirContentsToFileList(&dir);
				}
			}
		}
	} else if (project_info.haveProject) {
		RDirectory dir;
		long i = 0;
		project_info.ReadLock();
		while (project_info.GetDirectory(i++, &dir) == B_NO_ERROR)
			AddDirContentsToFileList(&dir);
		project_info.ReadUnlock();
	} else {
		RFile file(parentRef);
		RDirectory dir;
		file.GetParent(&dir);
		AddDirContentsToFileList(&dir);
	}
	list.SortItems(CompareItems);
}

///   CompareItems
//-----------------------------------------------------------------------------

int TFindInfo::CompareItems(const void *a, const void *b)
{
	return strcmp((**(TFindFileListElement**)a).name,
		(**(TFindFileListElement**)b).name);
}

///   AddDirContentsToFileList
//-----------------------------------------------------------------------------

void TFindInfo::AddDirContentsToFileList(RDirectory *dir)
{
	long i = 0;
	RFile file;
	
	while (dir->GetFile(i++, &file) == B_NO_ERROR) {
		char name[B_FILE_NAME_LENGTH+1];
		file.GetName(name);
		if (FileInScope(file, name)) {
			TFindFileListElement *el = new TFindFileListElement;
			strcpy(el->name, name);
			el->ref = file.Ref();
			list.AddItem(el);
		}
	}
}

///      FileInScope
//-----------------------------------------------------------------------------

bool TFindInfo::FileInScope(BFile &file, const char *name)
{
	long len = strlen(name);
	switch (scope) {
		case FIND_TEXT_FILES:
			ulong type, creator;
			file.GetTypeAndApp(&type, &creator);
			return type == 'TEXT';
		case FIND_CPP_FILES:
			return len > 4 && strcmp(name+len-4, ".cpp") == 0;
		case FIND_H_FILES:
		case FIND_BE_HEADER_FILES:
		case FIND_STD_C_HEADER_FILES:
			return len > 2 && strcmp(name+len-2, ".h") == 0;
		case FIND_CPP_AND_H_FILES:
			return 
				(len > 4 && strcmp(name+len-4, ".cpp") == 0) ||
				(len > 2 && strcmp(name+len-2, ".h") == 0);
	}
	return FALSE;
}
