//-----------------------------------------------------------------------------
//
// TProjectInfo.cpp
//
// Project information class. 
//
// Shared by threads. Must be locked whenever accessed via the "Lock" and 
// "Unlock" member functions. The app thread and window threads may obtain the // lock. To prevent deadlocks, while you have the lock, you must not lock the 
// app or any windows.
//
// An exception is the "haveProject" data member. This flag starts out FALSE,
// becomes TRUE as soon as a project directory is set, and then remains TRUE 
// throughout the remainder of the program's life. You do not need to obtain
// the lock, and should not obtain the lock, if all you want to do is query
// the "haveProject" flag.
//
// The variables shell, beIncludes, etc. are always available, even if
// "haveProject" is FALSE. They are initialized at program startup. You do 
// need to get the read lock before accessing them, however. The values 
// of these variables can change during the life of the program.
//
//-----------------------------------------------------------------------------

#include "TProjectInfo.h"
#include "RFile.h"
#include "RDirectory.h"
#include "TGlobals.h"
#include "RUtil.h"
#include "RWriteLocker.h"
#include "RAlert.h"

B_DEFINE_CLASS_INFO(TProjectInfo, BObject);

RWriteLocker TProjectInfo::lock;

///TProjectInfo
//-----------------------------------------------------------------------------

TProjectInfo::TProjectInfo()
	: haveProject(FALSE), shell(NULL), beIncludes(NULL), includes(NULL),
	obj(NULL), app(NULL), cc(NULL), cflags(NULL), ld(NULL), ldflags(NULL),
	beHeaders(NULL), numCompileThreads(0)
{
	// Read the "/system/project" file to set the initial values of
	// the info variables.
	
	record_ref systemRef;
	if (get_ref_for_path("/system", &systemRef) == B_NO_ERROR) {;
		RDirectory *system = new RDirectory(systemRef);
		if (!system->Error()) ReadProjectFile(system);
	}
	
	// Set the other variables if they haven't already 
	// been set.
	
	if (!shell) CopyString(&shell, "/bin/sh");
	if (!beIncludes)
		CopyString(&beIncludes,
			"-I /boot/be/Headers/inc"
			" -I /boot/be/Headers/app"
			" -I /boot/be/Headers/device"
			" -I /boot/be/Headers/interface"
			" -I /boot/be/Headers/media"
			" -I /boot/be/Headers/midi"
			" -I /boot/be/Headers/net"
			" -I /boot/be/Headers/kernel"
			" -I /boot/be/Headers/storage"
			" -I /boot/be/Headers/support");
	if (!includes) CopyString(&includes, "");
	if (!obj) CopyString(&obj, "obj");
	if (!app) CopyString(&app, "");
	if (!cc) CopyString(&cc, "/boot/be/Tools/mwcc");
	if (!cflags) CopyString(&cflags,
		"-nosyspath -DPPC603=1 -Hcpplvl=3 -I./");
	if (!ld) CopyString(&ld, "/boot/be/Tools/bld");
	if (!ldflags) CopyString(&ldflags,
		"-nodefaults /boot/be/Libraries/libbe.so /boot/be/Libraries/libdll.a");
	if (!beHeaders) CopyString(&beHeaders, "/boot/be/Headers");
	if (!numCompileThreads) numCompileThreads = 2;
}

///~TProjectInfo
//-----------------------------------------------------------------------------

TProjectInfo::~TProjectInfo()
{
	delete [] shell;
	delete [] beIncludes;
	delete [] includes;
	delete [] obj;
	delete [] app;
	delete [] cc;
	delete [] cflags;
	delete [] ld;
	delete [] ldflags;
	delete [] beHeaders;
	
	long i = 0;
	while (RDirectory *dir = (RDirectory*)dirs.ItemAt(i++)) delete dir;
	i = 0;
	while (RDirectory *objDir = (RDirectory*)objs.ItemAt(i++)) delete objDir;
}

///operator=
//-----------------------------------------------------------------------------

TProjectInfo &TProjectInfo::operator=(const TProjectInfo &rhs)
{
	if (this != &rhs) {
		haveProject = rhs.haveProject;
		CopyString(&shell, rhs.shell);
		CopyString(&beIncludes, rhs.beIncludes);
		CopyString(&includes, rhs.includes);
		CopyString(&obj, rhs.obj);
		CopyString(&app, rhs.app);
		CopyString(&cc, rhs.cc);
		CopyString(&cflags, rhs.cflags);
		CopyString(&ld, rhs.ld);
		CopyString(&ldflags, rhs.ldflags);
		CopyString(&beHeaders, rhs.beHeaders);
		numCompileThreads = rhs.numCompileThreads;
	
		long i = 0;
		while (RDirectory *dir = (RDirectory*)dirs.ItemAt(i++)) delete dir;
		dirs.MakeEmpty();
		i = 0;
		while (RDirectory *dir = (RDirectory*)rhs.dirs.ItemAt(i++)) {
			RDirectory *dirCopy = new RDirectory;
			*dirCopy = *dir;
			dirs.AddItem(dirCopy);
		}
		
		i = 0;
		while (RDirectory *objDir = (RDirectory*)objs.ItemAt(i++))
			 delete objDir;
		objs.MakeEmpty();
		i = 0;
		while (RDirectory *objDir = (RDirectory*)rhs.objs.ItemAt(i++)) {
			RDirectory *objDirCopy = new RDirectory;
			*objDirCopy = *objDir;
			objs.AddItem(objDirCopy);
		}
	} 
	return *this;
}

///ReadLock
//-----------------------------------------------------------------------------

void TProjectInfo::ReadLock()
{
	lock.ReadLock();
}

///ReadUnlock
//-----------------------------------------------------------------------------

void TProjectInfo::ReadUnlock()
{
	lock.ReadUnlock();
}

///WriteLock
//-----------------------------------------------------------------------------

void TProjectInfo::WriteLock()
{
	lock.WriteLock();
}

///WriteUnlock
//-----------------------------------------------------------------------------

void TProjectInfo::WriteUnlock()
{
	lock.WriteUnlock();
}

///AddObjDirs
//-----------------------------------------------------------------------------

long TProjectInfo::AddObjDirs()
{
	long i = 0;
	while (RDirectory *dir = (RDirectory*)dirs.ItemAt(i++)) {
		RDirectory *objDir = new RDirectory;
		if (dir->GetDirectory(obj, objDir) != B_NO_ERROR) {
			long err = dir->Create(obj, objDir);
			if (err) return err;
	 	}
		objs.AddItem(objDir);
	}
	return B_NO_ERROR;
}

///   MakeEmpty
//-----------------------------------------------------------------------------

void TProjectInfo::MakeEmpty()
{
	long j = 0;
	while (RDirectory *dir = (RDirectory*)dirs.ItemAt(j++))
		delete dir;
	dirs.MakeEmpty();
	j = 0;
	while (RDirectory *objDir = (RDirectory*)objs.ItemAt(j++))
		delete objDir;
	objs.MakeEmpty();
}

///GetDirectory
//-----------------------------------------------------------------------------

long TProjectInfo::GetDirectory(const long dirIndex, RDirectory *dir)
{
	RDirectory *d = (RDirectory*)dirs.ItemAt(dirIndex);
	if (d == NULL) return B_ERROR;
	*dir = *d;
	return B_NO_ERROR;
}

///GetObjDirectory
//-----------------------------------------------------------------------------

long TProjectInfo::GetObjDirectory(const long dirIndex, RDirectory *objDir)
{
	RDirectory *d = (RDirectory*)objs.ItemAt(dirIndex);
	if (d == NULL) return B_ERROR;
	*objDir = *d;
	return B_NO_ERROR;
}

///GetFile
//-----------------------------------------------------------------------------

long TProjectInfo::GetFile(const long index, RFile *file, long *dirIndex)
{
	long inx = index;
	long i = 0;
	while (RDirectory *dir = (RDirectory*)dirs.ItemAt(i++)) {
		if (dir->GetFile(inx, file) == B_NO_ERROR) {
			if (dirIndex) *dirIndex = i-1;
			return B_NO_ERROR;
		}
		inx -= dir->CountFiles();
	}
	return B_ERROR;
}

long TProjectInfo::GetFile(const char *name, RFile *file, long *dirIndex)
{
	long i = 0;
	while (RDirectory *dir = (RDirectory*)dirs.ItemAt(i++)) {
		if (dir->GetFile(name, file) == B_NO_ERROR) {
			if (dirIndex) *dirIndex = i-1;
			return B_NO_ERROR;
		}
	}
	return B_ERROR;
}

///GetPrimaryFile
//-----------------------------------------------------------------------------

long TProjectInfo::GetPrimaryFile(const char *name, RFile *file)
{
	RDirectory *dir = (RDirectory*)dirs.ItemAt(0);
	return dir->GetFile(name, file);
}
	

///GetObjFile
//-----------------------------------------------------------------------------

long TProjectInfo::GetObjFile(const long index, RFile *file, long *dirIndex)
{
	long inx = index;
	long i = 0;
	while (RDirectory *dir = (RDirectory*)objs.ItemAt(i++)) {
		if (dir->GetFile(inx, file) == B_NO_ERROR) {
			if (dirIndex) *dirIndex = i-1;
			return B_NO_ERROR;
		}
		inx -= dir->CountFiles();
	}
	return B_ERROR;
}

long TProjectInfo::GetObjFile(const char *name, RFile *file, long *dirIndex)
{
	long i = 0;
	while (RDirectory *dir = (RDirectory*)objs.ItemAt(i++)) {
		if (dir->GetFile(name, file) == B_NO_ERROR) {
			if (dirIndex) *dirIndex = i-1;
			return B_NO_ERROR;
		}
	}
	return B_ERROR;
}

///SetProjectInfo
//-----------------------------------------------------------------------------

bool TProjectInfo::SetProjectInfo(record_ref ref)
{
	// Sets the project info.
	//
	// Returns TRUE if project info set, FALSE if error.

	// Set the primary directory.

	RDirectory *primary = new RDirectory(ref);
	if (primary->Error()) {
		RAlert::ErrMsg("Could not set primary project directory.");
		delete primary;
		return FALSE;
	}
	dirs.AddItem(primary, 0);
	
	// Read the project file.

	ReadProjectFile(primary);
	
	// Set info.app to the name of the primary project directory 
	// if it hasn't already been set.
	
	if (!app || *app == 0) {
		app = new char[B_FILE_NAME_LENGTH+1];
		primary->GetName(app);
	}
	
	// Initialize the obj directories.
	
	if (AddObjDirs() != B_NO_ERROR) {
		RAlert::ErrMsg("Could not set project directory.");
		return FALSE;
	}
	
	haveProject = TRUE;
	
	return TRUE;
}

///   ReadProjectFile
//-----------------------------------------------------------------------------

void TProjectInfo::ReadProjectFile(RDirectory *dir)
{
	RFile file;
	
	if (dir->GetFile("project", &file) == B_NO_ERROR) {
		char *data = (char*)file.ReadToMem();
		if (file.Error()) return;
		ParseOneVariable(data, "SHELL", &shell);
		ParseOneVariable(data, "BE_INCLUDES", &beIncludes);
		ParseOneVariable(data, "INCLUDES", &includes);
		ParseOneVariable(data, "OBJ", &obj);
		ParseOneVariable(data, "APP", &app);
		ParseOneVariable(data, "CC", &cc);
		ParseOneVariable(data, "CFLAGS", &cflags);
		ParseOneVariable(data, "LD", &ld);
		ParseOneVariable(data, "LDFLAGS", &ldflags);
		ParseOneVariable(data, "BE_HEADERS", &beHeaders);
		char *numCompileThreadsStr = NULL;
		ParseOneVariable(data, "NUM_COMPILE_THREADS", &numCompileThreadsStr);
		if (numCompileThreadsStr) {
			numCompileThreads = atol(numCompileThreadsStr);
			if (numCompileThreads <= 0) numCompileThreads = 1;
			if (numCompileThreads > MAX_RUNNERS) 
				numCompileThreads = MAX_RUNNERS;
			delete [] numCompileThreadsStr;
		}
		ParseSecondaryDirectories(dir, data);
		delete [] data;
	}
}

///   ParseOneVariable
//-----------------------------------------------------------------------------

void TProjectInfo::ParseOneVariable(const char *data, const char *name, 
	char **v)
{
	long len = strlen(name);
	const char *p = data;
	while (*p) {
		while (*p == B_TAB || *p == ' ') p++;
		if (strncmp(p, name, len) == 0) {
			p += len;
			while (*p == B_TAB || *p == ' ') p++;
			if (*p == '=') {
				p++;
				while (*p == B_TAB || *p == ' ') p++;
				const char *q = p;
				while (*q && *q != B_RETURN) q++;
				delete [] *v;
				*v = new char[q-p+1];
				memcpy(*v, p, q-p);
				(*v)[q-p] = 0;
				return;
			}
		}
		while (*p && *p != B_RETURN) p++;
		if (!*p) break;
		p++;
	}
}

///   ParseSecondaryDirectories
//-----------------------------------------------------------------------------

void TProjectInfo::ParseSecondaryDirectories(RDirectory *dir, const char *data)
{
	const char name[] = "DIRECTORY";
	long len = strlen(name);
	char primaryPath[1000];
	dir->GetPath(primaryPath, 1000);
	if (dir->Error()) return;
	const char *p = data;
	while (*p) {
		while (*p == B_TAB || *p == ' ') p++;
		if (strncmp(p, name, len) == 0) {
			p += len;
			while (*p == B_TAB || *p == ' ') p++;
			if (*p == '=') {
				p++;
				while (*p == B_TAB || *p == ' ') p++;
				const char *q = p;
				while (*q && *q != B_RETURN) q++;
				if (q > p && q - p < 1000) {
					char val[1000];
					memcpy(val, p, q-p);
					val[q-p] = 0;
					char path[2000];
					if (*p == '/') {
						strcpy(path, val);
					} else {
						sprintf(path, "%s/%s", primaryPath, val);
					}
					record_ref ref;
					long err = get_ref_for_path(path, &ref);
					if (!err) {
						RDirectory *secondary = new RDirectory(ref);
						err = secondary->Error();
						if (!err) dirs.AddItem(secondary);
					}
					if (err) {
						char str[2200];
						sprintf(str, 
							"Could not add secondary directory %s.\n\n"
							"Directory skipped.", path);
						RAlert::InfoMsg(str);
					}
				}
			}
		}
		while (*p && *p != B_RETURN) p++;
		if (!*p) break;
		p++;
	}
}
