//-----------------------------------------------------------------------------
//
// TSearch.cpp
//
// Multi-file searching class.
//
// This global object runs in its own blooper thread to avoid deadlocks. When a 
// search operation running in a window thread cannot find any more matches 
// in its own window, if the user has specified a multi-file search, the code 
// in TTextView hands off the search to this thread to search more files.
//
// For replace all operations, if the search string is located in a new
// file, a REPLACE_ALL message is sent to the window containing the new
// match to continue the replace all operation in the new window.
//
// This class also contains a pair of static utility functions used both
// here and in TTextView. 
//
//-----------------------------------------------------------------------------

#include "TSearch.h"
#include "TFindInfo.h"
#include "RUtil.h"
#include "RTextView.h"
#include "TGlobals.h"
#include "RFile.h"
#include "TWindow.h"

B_DEFINE_CLASS_INFO(TSearch, BLooper);

///TSearch
//-----------------------------------------------------------------------------

TSearch::TSearch()
	: BLooper(B_NORMAL_PRIORITY)
{
	Run();
}

///SearchFileList
//-----------------------------------------------------------------------------

void TSearch::SearchFileList(bool replaceAll)
{
	BMessage *msg = new BMessage('doit');
	msg->AddBool("replaceAll", replaceAll);
	PostMessage(msg);
}

///MessageReceived
//-----------------------------------------------------------------------------

void TSearch::MessageReceived(BMessage *msg)
{
	if (msg->what == 'doit') {
		Doit(msg->FindBool("replaceAll"));
	} else {
		inherited::MessageReceived(msg);
	}
}

///Doit
//-----------------------------------------------------------------------------

void TSearch::Doit(bool replaceAll)
{
	bool matchFound = FALSE;
	TWindow *wind;
	
	while (!matchFound) {
	
		// Advance to the next file in the find file list.
	
		record_ref ref;
		find_info.Lock();
		TFindFileListElement *el = 
			(TFindFileListElement*)find_info.list.ItemAt(++find_info.pos);
		if (el) ref = el->ref;
		TFindScope scope = find_info.scope;
		find_info.Unlock();
		if (!el) break;
		
		// First check to see if the file is already open in one of our
		// windows. If it is, search the text in the window.
		
		bool windFound = FALSE;
		BList windowList = RWindow::WindowList();
		for (long i = 0; 
			wind = (TWindow*)windowList.ItemAt(i); 
			i++)
		{
			if (wind->Lock()) {
				record_ref r = wind->Ref();
				if (r.database == ref.database && 
					r.record == ref.record) 
				{
					windFound = TRUE;
					RTextView *textView = 
						(RTextView*)wind->FindView("textView");
					long size = textView->TextLength();
					long matchStart, matchFinish;
					find_info.Lock();
					matchFound = SearchText(textView->Text(), size, 0, size,
						&matchStart, &matchFinish);
					find_info.Unlock();
					if (matchFound) {
						wind->Activate();
						textView->Select(matchStart, matchFinish);
						textView->ScrollSelectionToMiddle();
					}
				}
				wind->Unlock();
			}
			if (windFound) break;
		}
		
		if (matchFound) break;
		if (windFound) continue;
		
		// If we couldn't find the file in one of our windows, open it
		// and search it. If we find a match, open the file in a new
		// window.
		
		RFile file(ref);
		long size;
		char *data = (char*)file.ReadToMem(&size);
		long matchStart, matchFinish;
		find_info.Lock();
		matchFound = SearchText(data, size, 0, size, &matchStart, &matchFinish);
		find_info.Unlock();
		if (matchFound) { 
			char title[B_FILE_NAME_LENGTH+1];
			file.GetName(title);
			wind = new TWindow(data, size, 
				scope == FIND_BE_HEADER_FILES || 
					scope == FIND_STD_C_HEADER_FILES, 
				ref.database, ref.record, title, matchStart, matchFinish);
			delete [] data;
		} else {
			delete [] data;
		}
	}
	
	if (matchFound) {
		if (replaceAll) {
			if (wind->Lock()) {
				wind->PostMessage(new BMessage(REPLACE_ALL),
					wind->PreferredReceiver());
				wind->Unlock();
			}
		}
	} else {
		beep();
	}
}

///BreaksAtChar
//-----------------------------------------------------------------------------

bool TSearch::BreaksAtChar(ulong aChar)
{
	return !(isalnum(aChar) | aChar == '_');
}

///SearchText
//-----------------------------------------------------------------------------

bool TSearch::SearchText(const char *text, long textLength, 
	long start, long finish, long *matchStart, long *matchFinish)
{
	// Caller must have find_info lock.

	// Searches from start to finish in text for a match with the
	// current search string. If one is found, return TRUE and the
	// offsets of the match in matchStart and matchFinish.
	
	char *searchString = find_info.searchString;
	if (searchString == NULL) return FALSE;
	long searchLen = strlen(searchString);
	if (searchLen == 0) return FALSE;
	finish -= searchLen;
	if (start > finish) return FALSE;
	const char *p, *q;
	if (find_info.forward) {
		if (find_info.caseSensitive) {
			for (p = text + start; p <= text + finish; p++) {
				if (strncmp(p, searchString, searchLen) == 0) {
					if (find_info.wholeWord) {
						if (p > text && !BreaksAtChar(*(p-1))) continue;
						q = p + searchLen;
						if (q < text + textLength && 
							!BreaksAtChar(*q)) continue;
					}
					break;
				}
			}
		} else {
			for (p = text + start; p <= text + finish; p++) {
				if (strncasecmp(p, searchString, searchLen) == 0) {
					if (find_info.wholeWord) {
						if (p > text && !BreaksAtChar(*(p-1))) continue;
						q = p + searchLen;
						if (q < text + textLength && 
							!BreaksAtChar(*q)) continue;
					}
					break;
				}
			}
		}
		if (p <= text + finish) {
			*matchStart = p-text;
			*matchFinish = p-text+searchLen;
			return TRUE;
		} else {
			return FALSE;
		}
	} else { // search backwards
		if (find_info.caseSensitive) {
			for (p = text + finish; p >= text + start; p--) {
				if (strncmp(p, searchString, searchLen) == 0) {
					if (find_info.wholeWord) {
						if (p > text && !BreaksAtChar(*(p-1))) continue;
						q = p + searchLen;
						if (q < text + textLength && 
							!BreaksAtChar(*q)) continue;
					}
					break;
				}
			}
		} else {
			for (p = text + finish; p >= text + start; p--) {
				if (strncasecmp(p, searchString, searchLen) == 0) {
					if (find_info.wholeWord) {
						if (p > text && !BreaksAtChar(*(p-1))) continue;
						q = p + searchLen;
						if (q < text + textLength && 
							!BreaksAtChar(*q)) continue;
					}
					break;
				}
			}
		}
		if (p >= text + start) {
			*matchStart = p-text;
			*matchFinish = p-text+searchLen;
			return TRUE;
		} else {
			return FALSE;
		}
	}
}
