//-----------------------------------------------------------------------------
//
// TTextView.cpp
//
// Text view class.
//
//-----------------------------------------------------------------------------

#include "TTextView.h"
#include "TWindow.h"
#include "TFindDlg.h"
#include "TGoToLineDlg.h"
#include "TSearch.h"
#include "RUtil.h"
#include "RAlert.h"
#include "TGlobals.h"
#include "TUndo.h"

B_DEFINE_CLASS_INFO(TTextView, RTextView);

///TTextView
//-----------------------------------------------------------------------------

TTextView::TTextView(BRect frame)
	: RTextView(frame, "textView", frame, B_FOLLOW_ALL, 
		B_WILL_DRAW | B_FRAME_EVENTS),
	undo(this)
{
	BRect textRect = frame;
	textRect.OffsetTo(0, 0);
	textRect.InsetBy(4, 4);
	SetTextRect(textRect);
}

///AttachedToWindow
//-----------------------------------------------------------------------------

void TTextView::AttachedToWindow()
{
	inherited::AttachedToWindow();
	MakeFocus();
	SetFontName("Kate");
	SetFontSize(9);
	SetWordWrap(TRUE);
	SetAutoindent(TRUE);
	float charWidth = StringWidth("A");
	SetTabWidth(4*charWidth);
	for (long aChar = 0; aChar < ' '; aChar++) {
		if (aChar != B_BACKSPACE && aChar != B_ENTER && aChar != B_TAB &&
			aChar != B_LEFT_ARROW && aChar != B_RIGHT_ARROW &&
			aChar != B_UP_ARROW && aChar != B_DOWN_ARROW &&
			aChar != B_HOME && aChar != B_END &&
			aChar != B_PAGE_UP && aChar != B_PAGE_DOWN)
		{
			DisallowChar(aChar);
		}
	}
}

///KeyDown
//-----------------------------------------------------------------------------

void TTextView::KeyDown(ulong aChar)
{
	if (IsEditable()) {
		bool dirty = TRUE;
		switch (aChar) {
			case B_BACKSPACE:
				long start, finish;
				GetSelection(&start, &finish);
				if (start == 0 && finish == 0) dirty = FALSE;
				break;
			case B_LEFT_ARROW:
			case B_RIGHT_ARROW:
			case B_UP_ARROW:
			case B_DOWN_ARROW:
			case B_HOME:
			case B_END:
			case B_PAGE_UP:
			case B_PAGE_DOWN:
				dirty = FALSE;
				break;
		}
		if (dirty) {
			undo.RecordKey(aChar);
			SetDirty();
		}
	}
	
	inherited::KeyDown(aChar);
}	

///FrameResized
//-----------------------------------------------------------------------------

void TTextView::FrameResized(float width, float height)
{
	BRect textRect(0, 0, width, height);
	textRect.InsetBy(4,4);
	SetTextRect(textRect);
	AdjustScrollBar();
	TWindow *wind = (TWindow*)Window();
	if (wind->scrollToMiddleAfterNextResizeHack) {
		ScrollSelectionToMiddle();
		wind->scrollToMiddleAfterNextResizeHack = FALSE;
	}
		
}

///BreaksAtChar
//-----------------------------------------------------------------------------

bool TTextView::BreaksAtChar(ulong aChar) const
{
	return TSearch::BreaksAtChar(aChar);
}

///Cut
//-----------------------------------------------------------------------------

void TTextView::Cut(BClipboard *clipboard)
{
	if (!IsEditable()) return;
	long start, finish;
	GetSelection(&start, &finish);
	if (start >= finish) return;
	undo.RecordCut();
	SetDirty();
	inherited::Cut(clipboard);
}

///Paste
//-----------------------------------------------------------------------------

void TTextView::Paste(BClipboard *clipboard)
{
	if (!IsEditable()) return;
	clipboard->Lock();
	long numBytes;
	const char *text = clipboard->FindText(&numBytes);
	bool clipboardHasText = text != NULL && numBytes > 0;
	clipboard->Unlock();
	if (!clipboardHasText) return;
	undo.RecordPaste(clipboard);
	SetDirty();
	inherited::Paste(clipboard);
}

///MessageReceived
//-----------------------------------------------------------------------------

void TTextView::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case UNDO:
			undo.Undo();
			break;
		case B_DELETE:
			if (!IsEditable()) return;
			long start, finish;
			GetSelection(&start, &finish);
			if (start >= finish) break;
			undo.RecordDelete();
			Delete();
			ScrollToSelection();
			SetDirty();
			break;
		case SELECT_ALL:
			Select(0, TextLength());
			break;
		case INDENT:
		case UN_INDENT:
		case COMMENT:
		case UN_COMMENT:
			if (!IsEditable()) return;
			IndentOrComment(msg->what);
			break;
		case MOVE_UP:
			MoveUp();
			break;
		case MOVE_DOWN:
			MoveDown();
			break;
		case FIND:
			new TFindDlg((TWindow*)Window());
			break;
		case FIND_AGAIN:
			find_info.Lock();
			FindAgain(FALSE);
			find_info.Unlock();
			break;
		case FIND_SELECTION:
			find_info.Lock();
			EnterSelection(&find_info.searchString);
			FindAgain(FALSE);
			find_info.Unlock();
			break;
		case FIND_IN_NEXT_FILE:
			search_blooper->SearchFileList(FALSE);
			break;
		case REPLACE:
			find_info.Lock();
			Replace();
			find_info.Unlock();
			break;
		case REPLACE_AND_FIND_NEXT:
			find_info.Lock();
			if (Replace()) FindAgain(FALSE);
			find_info.Unlock();
			break;
		case REPLACE_ALL:
			find_info.Lock();
			ReplaceAll();
			find_info.Unlock();
			undo.Clear();
			break;
		case ENTER_SEARCH_STRING:
			find_info.Lock();
			EnterSelection(&find_info.searchString);
			find_info.Unlock();
			break;
		case ENTER_REPLACE_STRING:
			find_info.Lock();
			EnterSelection(&find_info.replaceString);
			find_info.Unlock();
			break;
		case GO_TO_LINE_DIALOG:
			new TGoToLineDlg(Window(), this);
			break;
		case GO_TO_LINE_DOIT:
			bool emess = msg->HasString("emess1");
			long line = msg->FindLong("line");
			if (emess) 
				((TWindow*)Window())->DisplayCompilerErrorMessage(
					msg->FindString("emess1"),
					msg->FindString("emess2"),
					msg->FindString("emess3"));
			GoToLine(line, emess);
			break;
		case MAKE_C_STRING:
			MakeCString();
			break;
		case GO_TO_MARKER:
			GoToMarker(msg->FindString("theMarker"));
			break;
		default:
			inherited::MessageReceived(msg);
			break;
	}
}

///SetDirty
//-----------------------------------------------------------------------------

void TTextView::SetDirty()
{
	if (!IsEditable()) return;
	((TWindow*)Window())->SetDirty(TRUE);
}

///IndentOrComment
//-----------------------------------------------------------------------------

void TTextView::IndentOrComment(ulong what)
{
	if (!IsEditable()) return;
	long start, finish;
	GetExpandedSelection(&start, &finish);
	long len = TextLength();

	long selEnd = finish;
	
	if (CharAt(finish-1) == B_RETURN) finish--;
	finish--;
	while (finish >= start && CharAt(finish) != B_RETURN) finish--;
	finish++;

	// finish now points to first char of last line to be modified.
	
	if (finish < start) return;
	
	while (finish >= start) {
		switch (what) {
			case INDENT:
				Select(finish, finish);
				Insert("\t");
			selEnd++;
				break;
			case UN_INDENT:
				if (CharAt(finish) == B_TAB) {
					Select(finish, finish+1);
					Delete();
					selEnd--;
				}
				break;
			case COMMENT:
				Select(finish, finish);
				Insert("//");
				selEnd += 2;
				break;
			case UN_COMMENT:
				if (CharAt(finish) == '/' && CharAt(finish+1) == '/') {
					Select(finish, finish+2);
					Delete();
					selEnd -= 2;
				}
				break;
		}
		finish -= 2;
		if (finish < start) break;
		while (finish >= start && CharAt(finish) != B_RETURN) finish--;
		finish++;
	}
	Select(start, selEnd);
	undo.Clear();
	SetDirty();
}

///MoveUp
//-----------------------------------------------------------------------------

void TTextView::MoveUp()
{
	if (!IsEditable()) return;
	long start, finish;
	GetExpandedSelection(&start, &finish);
	const char *text = Text();	
	const char *p = text + start - 1;
	if (p < text) return;
	p--;
	while (p >= text && *p != B_RETURN) p--;
	p++;
	long moveTo = p - text;
	char *buf = new char[finish-start+1];
	memcpy(buf, text+start, finish-start);
	buf[finish-start] = 0;
	Select(start, finish);
	Delete();
	Select(moveTo, moveTo);
	Insert(buf);
	delete [] buf;
	Select(moveTo, moveTo+finish-start);
	undo.Clear();
	SetDirty();
}

///MoveDown
//-----------------------------------------------------------------------------

void TTextView::MoveDown()
{
	if (!IsEditable()) return;
	long start, finish;
	GetExpandedSelection(&start, &finish);
	const char *text = Text();	
	const char *p = text + finish;
	const char *pEnd = text + TextLength();
	while (p < pEnd && *p != B_RETURN) p++;
	if (p >= pEnd) return;
	p++;
	long moveTo = p - text;
	char *buf = new char[finish-start+1];
	memcpy(buf, text+start, finish-start);
	buf[finish-start] = 0;
	Select(start, finish);
	Delete();
	moveTo -= finish-start;
	Select(moveTo, moveTo);
	Insert(buf);
	delete [] buf;
	Select(moveTo, moveTo+finish-start);
	undo.Clear();
	SetDirty();
}

///GetExpandedSelection
//-----------------------------------------------------------------------------

void TTextView::GetExpandedSelection(long *start, long *finish)
{
	// Get current selection expanded on both ends to line boundaries.

	long s, f;
	GetSelection(&s, &f);
	long len = TextLength();
	s--;
	while (s >= 0 && CharAt(s) != B_RETURN) s--;
	s++;
	if (f > 0) f--;
	while (f < len && CharAt(f) != B_RETURN) f++;
	if (f < len) f++;
	*start = s;
	*finish = f;
}

///EnterSelection
//-----------------------------------------------------------------------------

void TTextView::EnterSelection(char **str)
{
	long start, finish, i;
	GetSelection(&start, &finish);
	if (start >= finish) return;
	for (i = start; i < finish; i++) if (CharAt(i) == B_RETURN) return;
	delete [] *str;
	char *s = new char [finish-start+1];
	char *p;
	for (i = start, p = s; i < finish; i++, p++)
		*p = CharAt(i);
	*p = 0;
	*str = s;
	
}

///AdjustScrollBar
//-----------------------------------------------------------------------------

void TTextView::AdjustScrollBar()
{
	BScrollView *scrollView =(BScrollView*)Parent();
	long height = Frame().bottom - Frame().top;
	BScrollBar *vScroll = scrollView->ScrollBar(B_VERTICAL);
	long lineHeight, bigStep;
	vScroll->GetSteps(&lineHeight, &bigStep);
	bigStep = height - lineHeight;
	if (bigStep < lineHeight) bigStep = lineHeight;
	vScroll->SetSteps(lineHeight, bigStep); 
}

///FindAgain
//-----------------------------------------------------------------------------

bool TTextView::FindAgain(bool replaceAll)
{
	// Caller must have find_info lock.

	bool result;
	if (find_info.scope == FIND_THIS_FILE || find_info.pos >= 0) {
		long start, finish, textLength;
		GetSelection(&start, &finish);
		textLength = TextLength();
		if (find_info.forward) {
			result = SearchCurrentFile(finish, textLength);
			if (!result) {
				if (find_info.scope == FIND_THIS_FILE) {
					if (find_info.wrap)
						result = SearchCurrentFile(0, finish);
				} else {
					search_blooper->SearchFileList(replaceAll);
					return FALSE;
				}
			}
		} else {
			result = SearchCurrentFile(0, start);
			if (!result) {
				if (find_info.scope == FIND_THIS_FILE) {
					if (find_info.wrap)
						result = SearchCurrentFile(start, textLength);
				} else {
					search_blooper->SearchFileList(replaceAll);
					return FALSE;
				}
			}
		}
	} else {
		search_blooper->SearchFileList(replaceAll);
		return FALSE;
	}
	if (!result) beep();
	return result;
}

///SearchCurrentFile
//-----------------------------------------------------------------------------

bool TTextView::SearchCurrentFile(long start, long finish)
{
	// Caller must have find_info lock.

	long matchStart, matchFinish;
	if (TSearch::SearchText(Text(), TextLength(), start, finish, 
		&matchStart, &matchFinish))
	{
		Select(matchStart, matchFinish);
		ScrollSelectionToMiddle();
		return TRUE;
	} else {
		return FALSE;
	}
}

///CanReplace
//-----------------------------------------------------------------------------

bool TTextView::CanReplace()
{
	// Caller must have find_info lock.

	if (!IsEditable()) return FALSE;
	char *searchString = find_info.searchString;
	long searchLen = strlen(searchString);
	if (searchLen == 0) return FALSE;
	long start, finish;
	GetSelection(&start, &finish);
	if (searchLen != finish-start) return FALSE;
	long i;
	char *p;
	if (find_info.caseSensitive) {
		for (i = start, p = searchString; *p; i++, p++)
			if (CharAt(i) != *p) return FALSE;
	} else {
		for (i = start, p = searchString; *p; i++, p++)
			if (toupper(CharAt(i)) != toupper(*p)) return FALSE;
	}
	if (find_info.wholeWord) {
		if (start > 0 && !TSearch::BreaksAtChar(CharAt(start-1))) return FALSE;
		long textLength = TextLength();
		if (finish < textLength &&
			!TSearch::BreaksAtChar(CharAt(finish))) return FALSE;
	}
	return TRUE;
}

///Replace
//-----------------------------------------------------------------------------

bool TTextView::Replace()
{
	// Caller must have find_info lock.

	if (!CanReplace()) return FALSE;
	undo.RecordReplace();
	Delete();
	Insert(find_info.replaceString);
	if (!find_info.forward) {
		long replaceLen = strlen(find_info.replaceString);
		long start, finish;
		GetSelection(&start, &finish);
		Select(start-replaceLen, finish-replaceLen);
	}
	ScrollSelectionToMiddle();
	SetDirty();
	return TRUE;
}

///ReplaceAll
//-----------------------------------------------------------------------------

void TTextView::ReplaceAll()
{
	// Caller must have find_info lock.

	while (TRUE) {
		if (!Replace()) return;
		Window()->UpdateIfNeeded();
		if (!FindAgain(TRUE)) return;
	}
}

///GoToLine
//-----------------------------------------------------------------------------

void TTextView::GoToLine(long line, bool emess)
{
	line--;
	
	const char *text = Text();
	const char *p, *q;
	long n = 0;
	for (p = text; *p; p++) {
		if (n >= line) break;
		if (*p == B_RETURN) n++;
	}
	long start, finish;
	if (*p) {
		for (q = p; *q && *q != B_RETURN; q++);
		start = p - text;
		finish = q - text + 1;
	} else {
		start = finish = TextLength();
	}
	Select(start, finish);
	ScrollSelectionTo(emess ? 0.1 : 0.5, emess ? TRUE : FALSE);
}

///GoToMarker
//-----------------------------------------------------------------------------

void TTextView::GoToMarker(const char *marker)
{
	const char *text = Text();
	for (const char *t = text; *t; t++) {
		if (*t == '/' && *(t+1) == '/' && *(t+2) == '/') {
			char name[100];
			const char *p;
			char *q;
			t += 3;
			for (p = t, q = name; 
				*p && *p != B_RETURN && q < name+99; 
				p++, q++) *q = *p;
			*q = 0;
			if (strcmp(marker, name) == 0) {
				Select(t-text, p-text);
				ScrollSelectionToTop();
				break;
			}
		}
		while (*t && *t != B_RETURN) t++;
	}
}

///MakeCString
//-----------------------------------------------------------------------------

void TTextView::MakeCString()
{
	long start, finish;
	GetSelection(&start, &finish);
	if (start >= finish) return;
	Select(start, start);
	Insert("\"");
	finish++;
	long i = start+1;
	while (i < finish) {
		switch (CharAt(i)) {
			case B_RETURN:
				Select(i, i+1);
				Delete();
				Insert("\\n");
				i += 2;
				finish++;
				break;
			case B_TAB:
				Select(i, i+1);
				Delete();
				Insert("\\t");
				i += 2;
				finish++;
				break;
			case '"':
				Select(i, i+1);
				Delete();
				Insert("\\\"");
				i += 2;
				finish++;
				break;
			case '\\':
				Select(i, i+1);
				Delete();
				Insert("\\\\");
				i += 2;
				finish++;
				break;
			default:
				i++;
		}
	}
	Select(i, i);
	Insert("\"");
	finish++;
	Select(start, finish);
	ScrollSelectionToMiddle();
	undo.Clear();
	SetDirty();
}