//-----------------------------------------------------------------------------
//
// TWindow.cpp
//
// Edit window class.
//
//-----------------------------------------------------------------------------

#include "TWindow.h"
#include "TTextView.h"
#include "TGlobals.h"
#include "TFindInfo.h"
#include "RWindowsMenu.h"
#include "RAlert.h"
#include "TDepend.h"
#include "TUtil.h"
#include "RGlobals.h"
#include "TCompile.h"
#include "RFile.h"
#include "RDirectory.h"
#include "RPanelInfo.h"
#include "TProjectInfo.h"

B_DEFINE_CLASS_INFO(TWindow, RWindow);

long TWindow::untitledCount = 0;

///TPanelView
//-----------------------------------------------------------------------------

class TPanelView : public BView
{
	B_DECLARE_CLASS_INFO(BView);
public:
	TPanelView(BRect frame);
	void MouseDown(BPoint point);
};

B_DEFINE_CLASS_INFO(TPanelView, BView);

TPanelView::TPanelView(BRect frame)
	: BView(frame, "panelView", B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP, 
		B_WILL_DRAW)
{
}

void TPanelView::MouseDown(BPoint point)
{
	BRect frame = Frame();
	Hide();
	TWindow *wind = (TWindow*)Window();
	BView *scrollView = wind->FindView("scrollView");
	BRect scrollFrame = scrollView->Frame();
	long delta = scrollFrame.top - frame.top;
	scrollView->MoveBy(0, -delta);
	scrollView->ResizeBy(0, delta);
}

///TPanelStringView
//-----------------------------------------------------------------------------

class TPanelStringView : public BStringView
{
	B_DECLARE_CLASS_INFO(BStringView);
public:
	TPanelStringView(BRect frame, const char *name);
private:
	void MouseDown(BPoint point);
};

B_DEFINE_CLASS_INFO(TPanelStringView, BStringView);

TPanelStringView::TPanelStringView(BRect frame, const char *name)
	: BStringView(frame, name, "Hello, world", 
		B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP, 
		B_WILL_DRAW)
{
}

void TPanelStringView::MouseDown(BPoint point)
{
	TPanelView *panelView = (TPanelView*)Parent();
	panelView->MouseDown(point);
}

///TWindow
//-----------------------------------------------------------------------------
	
TWindow::TWindow(const void *data, const long size, const bool readonly,
	const database_id database, const record_id record, char *title,
	long start, long finish)
	: RWindow(BRect(0, 0, 200, 200), "", B_TITLED_WINDOW, 0),
	dirty(FALSE), readonly(readonly),
	scrollToMiddleAfterNextResizeHack(FALSE)
{
	Lock();
	
	SetSizeLimits(100.0, 1000000.0, 100.0, 1000000.0);	
	BRect bounds = Bounds();
	
	// Set the window title.
	
	ref.database = database;
	ref.record = record;
	
	if (title) {
		SetTitle(title);
	} else {
		char title[B_FILE_NAME_LENGTH+1];
		untitledCount++;
		if (untitledCount == 1) {
			strcpy(title, "untitled");
		} else {
			sprintf(title, "untitled %ld", untitledCount);
		}
		SetTitle(title);
	}

	// Create the menu bar.
	
	BRect menuBarFrame;
	CreateMenuBar(menuBarFrame);
	
	// Create the compiler error message panel view.
	
	CreatePanelView(menuBarFrame);
	
	// Create the text and scroll views.
	
	TTextView *text = CreateScrollView(menuBarFrame);
	
	// Apply the window staggering algorithm to set the window's
	// initial position and size.
	
	Stagger(text);
	
	// Create the initial windows menu for this window, and add this 
	// new window to all the other window's windows menus.
	
	windows_menu_blooper->Init(this);
	windows_menu_blooper->Add(this, FALSE);
	
	// If initial data, add it to the text view.
	
	if (data != NULL) {
		text->SetText((char*)data, size);
		BuildMarkersMenu();
		text->Select(start, finish);
		if (start > 0) scrollToMiddleAfterNextResizeHack = TRUE;
	}
	
	// If a file is specified, and it's name ends in .cpp or .h, and
	// if we don't already have a project directory, set the project directory.
	
	if (!readonly && !project_info.haveProject) {
		if (ref.database >= 0) {
			BFile file(ref);
			SetProjectDir(file);
		}
	}
			
	Show();	
	Unlock();
}

///CreateMenuBar
//-----------------------------------------------------------------------------

void TWindow::CreateMenuBar(BRect &menuBarFrame)
{
	menuBarFrame = Bounds();
	menuBarFrame.bottom = menuBarFrame.top + 14;
	menuBarFrame.right += 1;
	BMenuBar *bar = new BMenuBar(menuBarFrame, "menuBar");
	
	BMenu *menu;
	BMenuItem *item;
	
	menu = new BMenu("File");
	item = new BMenuItem("Save", new BMessage(SAVE), 'S');
	menu->AddItem(item);
	item = new BMenuItem("Save As...", new BMessage(SAVE_AS));
	menu->AddItem(item);
	item = new BMenuItem("Save a Copy...", new BMessage(SAVE_A_COPY));
	menu->AddItem(item);
	item = new BMenuItem("Revert to Saved", new BMessage(REVERT_TO_SAVED));
	menu->AddItem(item);
	menu->AddSeparatorItem();
	item = new BMenuItem("Open Partner (Alt-tab)", new BMessage(OPEN_PARTNER));
	menu->AddItem(item);
	item = new BMenuItem("Get Info...", new BMessage(GET_INFO), 'I');
	menu->AddItem(item);
	item = new BMenuItem("Convert to Script File", 
		new BMessage(CONVERT_TO_SCRIPT_FILE));
	menu->AddItem(item);
	item = new BMenuItem("Compile", new BMessage(COMPILE), 'K');
	menu->AddItem(item);
	menu->AddSeparatorItem();
	item = new BMenuItem("Close", new BMessage(B_QUIT_REQUESTED), 'W');
	menu->AddItem(item);
	bar->AddItem(menu);
	
	menu = new BMenu("Edit");
	item = new BMenuItem("Undo", new BMessage(UNDO), 'Z');
	menu->AddItem(item);
	menu->AddSeparatorItem();
	item = new BMenuItem("Cut", new BMessage(B_CUT), 'X');
	menu->AddItem(item);
	item = new BMenuItem("Copy", new BMessage(B_COPY), 'C');
	menu->AddItem(item);
	item = new BMenuItem("Paste", new BMessage(B_PASTE), 'V');
	menu->AddItem(item);
	item = new BMenuItem("Delete", new BMessage(B_DELETE));
	menu->AddItem(item);
	item = new BMenuItem("Select All", new BMessage(SELECT_ALL), 'A');
	menu->AddItem(item);
	menu->AddSeparatorItem();
	item = new BMenuItem("Indent", new BMessage(INDENT), ']');
	menu->AddItem(item);
	item = new BMenuItem("Un-Indent", new BMessage(UN_INDENT), '[');
	menu->AddItem(item);
	item = new BMenuItem("Comment", new BMessage(COMMENT), 'C', B_SHIFT_KEY);
	menu->AddItem(item);
	item = new BMenuItem("Un-Comment", new BMessage(UN_COMMENT), 
		'U', B_SHIFT_KEY);
	menu->AddItem(item);
	menu->AddSeparatorItem();
	item = new BMenuItem("Move Up (Alt-up)", new BMessage(MOVE_UP));
	menu->AddItem(item);
	item = new BMenuItem("Move Down (Alt-down)", new BMessage(MOVE_DOWN));
	menu->AddItem(item);
	menu->SetTarget(NULL, this);
	bar->AddItem(menu);
	
	menu = new BMenu("Search");
	item = new BMenuItem("Find...", new BMessage(FIND), 'F');
	menu->AddItem(item);
	item = new BMenuItem("Find Again", new BMessage(FIND_AGAIN), 'G');
	menu->AddItem(item);
	item = new BMenuItem("Find Selection", new BMessage(FIND_SELECTION), 'H');
	menu->AddItem(item);
	item = new BMenuItem("Find in Next File", 
		new BMessage(FIND_IN_NEXT_FILE), 'T');
	menu->AddItem(item);
	menu->AddSeparatorItem();
	item = new BMenuItem("Replace", new BMessage(REPLACE), '=');
	menu->AddItem(item);
	item = new BMenuItem("Replace & Find Next", 
		new BMessage(REPLACE_AND_FIND_NEXT), 'L');
	menu->AddItem(item);
	item = new BMenuItem("Replace All", new BMessage(REPLACE_ALL), 
		'L', B_SHIFT_KEY);
	menu->AddItem(item);
	menu->AddSeparatorItem();
	item = new BMenuItem("Enter Search String", 
		new BMessage(ENTER_SEARCH_STRING), 'E');
	menu->AddItem(item);
	item = new BMenuItem("Enter Replace String", 
		new BMessage(ENTER_REPLACE_STRING), 'Y');
	menu->AddItem(item);
	menu->AddSeparatorItem();
	item = new BMenuItem("Go To Line...", new BMessage(GO_TO_LINE_DIALOG), ',');
	menu->AddItem(item);
	menu->SetTarget(NULL, this);
	bar->AddItem(menu);

	menu = new BMenu("Geek");
	item = new BMenuItem("Make C String", new BMessage(MAKE_C_STRING), '1');
	menu->AddItem(item);
	menu->SetTarget(NULL, this);
	bar->AddItem(menu);
	
	menu = new BMenu("Markers");
	menu->AddSeparatorItem();
	item = new BMenuItem("How to Make Markers...", 
		new BMessage(HOW_TO_MAKE_MARKERS));
	item->SetTarget(this);
	menu->AddItem(item);
	bar->AddItem(menu);
	
	menu = new BMenu("Windows");
	bar->AddItem(menu);
	
	AddChild(bar);
	
	// Add the special key shortcuts.
	
	AddShortcut(B_TAB, 0, new BMessage(OPEN_PARTNER));
	AddShortcut(B_UP_ARROW, 0, new BMessage(MOVE_UP));
	AddShortcut(B_DOWN_ARROW, 0, new BMessage(MOVE_DOWN));
}

///CreatePanelView
//-----------------------------------------------------------------------------

void TWindow::CreatePanelView(BRect &menuBarFrame)
{	
	BRect bounds = Bounds();
	BRect panelFrame(bounds.left-1, menuBarFrame.bottom+1,
		bounds.right+1, menuBarFrame.bottom+44);
	TPanelView *panelView = new TPanelView(panelFrame);
	AddChild(panelView);
	panelView->SetViewColor(light_gray_background);
	
	BRect dividerFrame = panelView->Bounds();
	dividerFrame.top = dividerFrame.bottom-2;
	BBox *divider = new BBox(dividerFrame, "divider", 
		B_FOLLOW_LEFT_RIGHT | B_FOLLOW_BOTTOM, 
		B_WILL_DRAW);
	panelView->AddChild(divider);
	
	BRect emsgFrame = panelView->Bounds();
	emsgFrame.InsetBy(5, 3);
	emsgFrame.bottom = emsgFrame.top + 12;
	TPanelStringView *emsg = new TPanelStringView(emsgFrame, "emess1");
	panelView->AddChild(emsg);
	emsg->SetLowColor(light_gray_background);
	emsg->SetFontName("Kate");
	
	emsgFrame.OffsetBy(0, 12);
	emsg = new TPanelStringView(emsgFrame, "emess2");
	panelView->AddChild(emsg);
	emsg->SetLowColor(light_gray_background);
	emsg->SetFontName("Kate");
	
	emsgFrame.OffsetBy(0, 12);
	emsg = new TPanelStringView(emsgFrame, "emess3");
	panelView->AddChild(emsg);
	emsg->SetLowColor(light_gray_background);
	emsg->SetFontName("Kate");

	panelView->Hide();
}

///Stagger
//-----------------------------------------------------------------------------

void TWindow::Stagger(TTextView *text)
{
	float charWidth = text->StringWidth("A");
	float maxWidth = 80*charWidth + 12 + 4 + 4;
	screen_info screenInfo;
	get_screen_info(&screenInfo);
	
	BRect windFrame = screenInfo.frame;
	windFrame.top += 30;
	windFrame.left += 80;
	windFrame.bottom -= 10;
	windFrame.right -= 10;
	
	RWindow *front = Front();
	if (front && front != this && front->Lock()) {
		float frontFrameLeft = front->Frame().left;
		if (frontFrameLeft + 20 + maxWidth <= windFrame.right)
			windFrame.left = frontFrameLeft + 20;
		front->Unlock();
	}
	
	if (windFrame.right - windFrame.left > maxWidth) 
		windFrame.right = windFrame.left + maxWidth;
	MoveTo(windFrame.left, windFrame.top);
	ResizeTo(windFrame.right - windFrame.left, 
		windFrame.bottom - windFrame.top);
}

///CreateScrollView
//-----------------------------------------------------------------------------

TTextView *TWindow::CreateScrollView(BRect &menuBarFrame)
{
	BRect bounds = Bounds();
	BRect textFrame = bounds;
	textFrame.top = menuBarFrame.bottom+1;
	textFrame.right -= 12;
	TTextView *text = new TTextView(textFrame);
	if (readonly) text->MakeEditable(FALSE);
	
	BScrollView *scroll = new BScrollView("scrollView", text, B_FOLLOW_ALL, 0, 
		FALSE, TRUE, FALSE);
	AddChild(scroll);

	BScrollBar *vScroll = scroll->ScrollBar(B_VERTICAL);
	float lineHeight = text->LineHeight();
	vScroll->SetSteps(lineHeight, bounds.bottom - lineHeight);
	vScroll->SetRange(0, 0);
	return text;
}

///SetProjectDir
//-----------------------------------------------------------------------------

void TWindow::SetProjectDir(BFile &file)
{
	if (IsDotCpp(file) || IsDotH(file)) {
		RDirectory dir;
		file.GetParent(&dir);
		if (!file.Error()) depend_blooper->SetProject(dir.Ref());
	}
}

///QuitRequested
//-----------------------------------------------------------------------------

bool TWindow::QuitRequested()
{
	// If the window is dirty, present the "Save changes" alert.

	if (dirty) {
		char text[256];
		sprintf(text, "\nSave changes to \"%s\"?\n ", Title());
		RAlert *alert = new RAlert(text, Frame(), "Don't Save", "Cancel", 
			"Save", B_STOP_ALERT);
		alert->SetShortcut(0, 'd');
		alert->SetShortcut(1, B_ESCAPE);
		long result = alert->Go();
		switch (result) {
			case 0: /* Don't Save */
				break;
			case 1: /* Cancel */
				return FALSE;
			case 2: /* Save */
				if (ref.database >= 0) {
					RFile file(ref);
					SaveTextToFile(file);
				} else {
					Save(SAVE);
					return FALSE;
				}
				break;
		}
	}
	
	// If we're the last window, tell the app to quit.
	
	if (RWindow::CountWindows() <= 1) {
		be_app->PostMessage(B_QUIT_REQUESTED);
		return TRUE;
	}
	
	// Remove self from all other window's windows menus.
	
	windows_menu_blooper->Remove(this);
	return TRUE;
}

///MenusWillShow
//-----------------------------------------------------------------------------

void TWindow::MenusWillShow()
{
	BMenuBar *bar = (BMenuBar*)FindView("menuBar");
	BMenu *menu;
	BMenuItem *item;
	
	bool saved = ref.database >= 0;
	bool isDotCpp = IsDotCpp(ref);
	bool isDotH = IsDotH(ref);
	TTextView *textView = (TTextView*)FindView("textView");
	long start, finish;
	textView->GetSelection(&start, &finish);
	bool textSelected = start < finish;
	bool neverSavedButNotEmpty = 
		(ref.database < 0) && textView->TextLength() > 0;
	find_info.Lock();
	bool multiFileSearch = find_info.scope != FIND_THIS_FILE;
	bool canReplace = textView->CanReplace();
	find_info.Unlock();
	
	menu = bar->FindItem("File")->Submenu();
	item = menu->FindItem(SAVE);
	item->SetEnabled(dirty || neverSavedButNotEmpty);
	item = menu->FindItem(REVERT_TO_SAVED);
	item->SetEnabled(dirty && saved);
	item = menu->FindItem(OPEN_PARTNER);
	item->SetEnabled(isDotCpp || isDotH);
	item = menu->FindItem(CONVERT_TO_SCRIPT_FILE);
	item->SetEnabled(saved);
	item = menu->FindItem(COMPILE);
	item->SetEnabled((isDotCpp || isDotH) && saved && 
		project_info.haveProject);
	
	menu = bar->FindItem("Edit")->Submenu();
	item = menu->FindItem(UNDO);
	item->SetEnabled(textView->CanUndo());
	item = menu->FindItem(B_CUT);
	item->SetEnabled(!readonly && textSelected);
	item = menu->FindItem(B_COPY);
	item->SetEnabled(textSelected);
	item = menu->FindItem(B_PASTE);
	item->SetEnabled(!readonly);
	item = menu->FindItem(B_DELETE);
	item->SetEnabled(!readonly && textSelected);
	item = menu->FindItem(INDENT);
	item->SetEnabled(!readonly);
	item = menu->FindItem(UN_INDENT);
	item->SetEnabled(!readonly);
	item = menu->FindItem(COMMENT);
	item->SetEnabled(!readonly);
	item = menu->FindItem(UN_COMMENT);
	item->SetEnabled(!readonly);
	item = menu->FindItem(MOVE_UP);
	item->SetEnabled(!readonly);
	item = menu->FindItem(MOVE_DOWN);
	item->SetEnabled(!readonly);
	
	menu = bar->FindItem("Search")->Submenu();
	item = menu->FindItem(FIND_SELECTION);
	item->SetEnabled(textSelected);
	item = menu->FindItem(FIND_IN_NEXT_FILE);
	item->SetEnabled(multiFileSearch);
	item = menu->FindItem(REPLACE);
	item->SetEnabled(!readonly && canReplace);
	item = menu->FindItem(REPLACE_AND_FIND_NEXT);
	item->SetEnabled(!readonly && canReplace);
	item = menu->FindItem(REPLACE_ALL);
	item->SetEnabled(!readonly && canReplace);
	item = menu->FindItem(ENTER_SEARCH_STRING);
	item->SetEnabled(textSelected);
	item = menu->FindItem(ENTER_REPLACE_STRING);
	item->SetEnabled(textSelected);
	
	menu = bar->FindItem("Geek")->Submenu();
	item = menu->FindItem(MAKE_C_STRING);
	item->SetEnabled(!readonly && textSelected);
}

///MessageReceived
//-----------------------------------------------------------------------------

void TWindow::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case SAVE:
		case SAVE_AS:
		case SAVE_A_COPY:
			Save(msg->what);
			break;
		case REVERT_TO_SAVED:
			RevertToSaved();
			break;
		case OPEN_PARTNER:
			OpenPartner();
			break;
		case GET_INFO:
			GetInfo();
			break;
		case CONVERT_TO_SCRIPT_FILE:
			{
				RFile file(ref);
				if (!file.Error()) file.SetTypeAndApp('TEXT', 'SCPT');
			}
			break;
		case COMPILE:
			if (IsDotH(ref)) {
				RFile file(ref);
				if (!file.Error()) {
					BDirectory dir;
					if (file.GetParent(&dir) == B_NO_ERROR) {
						char name[B_FILE_NAME_LENGTH+1];
						file.GetName(name);
						long len = strlen(name);
						if (len+4 < B_FILE_NAME_LENGTH) {
							strcpy(name+len-2, ".cpp");
							if (dir.GetFile(name, &file) == B_NO_ERROR) {
								compile_blooper->Compile(file.Ref());
								be_app->PostMessage(
									new BMessage(HIDE_ALL_PANELS));
							}
						}
					}
				}
			} else {
				compile_blooper->Compile(ref);
				be_app->PostMessage(new BMessage(HIDE_ALL_PANELS));
			}
			break;
		case MOVE_UP:
			{
				TTextView *textView = (TTextView*)FindView("textView");
				textView->MoveUp();
			}
			break;
		case MOVE_DOWN:
			{
				TTextView *textView = (TTextView*)FindView("textView");
				textView->MoveDown();
			}
			break;
		case HOW_TO_MAKE_MARKERS:
			char str[1000] = 
				"To make a marker in your text, type the "
				"following comment on a line by itself:\n\n"
				"///marker_name\n\n"
				"Note the three slashes.\n\n"
				"To update the Markers menu, save the window.";
			RAlert::InfoMsg(str, Frame());
			break;
		default:
			inherited::MessageReceived(msg);
			break;
	}
}

///SavePanelClosed
//-----------------------------------------------------------------------------

void TWindow::SavePanelClosed(BMessage *msg)
{
	panel_info.Lock();
	panel_info.frame = msg->FindRect("frame");
	panel_info.dir = msg->FindRef("directory");
	panel_info.Unlock();
}

///SaveRequested
//-----------------------------------------------------------------------------

void TWindow::SaveRequested(record_ref directory, const char *fileName)
{
	BDirectory dir(directory);
	RFile file;
	BMessage *msg = CurrentMessage();
	
	// Create the file if necessary.
	
	if (dir.GetFile(fileName, &file) != B_NO_ERROR) {
		if (dir.Create(fileName, &file) != B_NO_ERROR) return;
		file.SetTypeAndApp('TEXT', SIGNATURE);
	}
	
	// Save the text to the file.
	
	SaveTextToFile(file);
	
	// If this isn't a "Save a Copy" operation, update the file
	// ref for this window, clear our dirty flag, and change the
	// window title to the file name, which may have changed.
	
	if (msg && !msg->HasBool("saveCopy")) {
	
		ref = file.Ref();
		
		if (strcmp(Title(), fileName) != 0) {
			SetTitle(fileName);
			windows_menu_blooper->Remove(this);
			windows_menu_blooper->Add(this, FALSE);
		}

		SetDirty(FALSE);
	}
}

///SaveTextToFile
//-----------------------------------------------------------------------------

void TWindow::SaveTextToFile(RFile &file)
{
	if (readonly) return;
	
	// Write the file.
	
	BTextView *textView = (BTextView*)FindView("textView");
	const char *text = textView->Text();
	long len = textView->TextLength();
	file.WriteFromMem(text, len);	
	if (file.Error()) {
		char name[B_FILE_NAME_LENGTH+1];
		file.GetName(name);
		char str[256];
		sprintf (str, "Could not write file %s.", name);
		RAlert::ErrMsg(str);
		return;
	}
	
	// If we don't have a project yet, call SetProjectDir to set it
	// if we're saving a .cpp or .h file.
	
	if (!project_info.haveProject) SetProjectDir(file);
	
	// If we're saving a "project" file, reinitialize the project info. 
	// If we have a project, and if we're saving a .cpp or .h file, update 
	// the include file dependencies for the file being saved.
	
	char name[B_FILE_NAME_LENGTH+1];
	file.GetName(name);
	RFile f;
	project_info.ReadLock();
	if (strcmp(name, "project") == 0) {
		RDirectory primary;
		if (project_info.haveProject &&
			project_info.GetDirectory(0, &primary) == B_NO_ERROR) 
		{
			depend_blooper->SetProject(primary.Ref());
		} else {
			depend_blooper->SetProject();
		}
	} else if (project_info.haveProject &&
		project_info.GetFile(name, &f) == B_NO_ERROR &&
		(IsDotCpp(name) || IsDotH(name)))
	{
		depend_blooper->UpdateIncludeFileDependencies(file.Ref(), 
			text, len);
	}
	project_info.ReadUnlock();
	
	// Rebuild the markers menu.
	
	BuildMarkersMenu();
}

///BuildMarkersMenu
//-----------------------------------------------------------------------------

void TWindow::BuildMarkersMenu()
{
	BTextView *textView = (BTextView*)FindView("textView");
	BMenuBar *bar = (BMenuBar*)FindView("menuBar");
	BMenu *menu = bar->FindItem("Markers")->Submenu();
	BMenuItem *item;
	
	long numItems = menu->CountItems();
	numItems -= 2;
	while (numItems--) menu->RemoveItem(menu->ItemAt(0));
	long index = 0;
	for (const char *text = textView->Text(); *text; text++) {
		if (*text == '/' && *(text+1) == '/' && *(text+2) == '/') {
			char name[100];
			const char *p;
			char *q;
			text += 3;
			for (p = text, q = name; 
				*p && *p != B_RETURN && q < name+99; 
				p++, q++) *q = *p;
			*q = 0;
			BMessage *msg = new BMessage(GO_TO_MARKER);
			msg->AddString("theMarker", name);
			item = new BMenuItem(name, msg);
			item->SetTarget(NULL, this);
			menu->AddItem(item, index++);
		}
		while (*text && *text != B_RETURN) text++;
	}
}

///RevertToSaved
//-----------------------------------------------------------------------------

void TWindow::RevertToSaved()
{
	if (!dirty) return;
	if (ref.database < 0) return;
	BTextView *textView = (BTextView*)FindView("textView");
	RFile file(ref);
	long size;
	char *data = (char*)file.ReadToMem(&size);
	textView->SetText(data, size);
	textView->Select(0, 0);
	textView->ScrollToSelection();
	delete [] data;
	SetDirty(FALSE);
	BuildMarkersMenu();
}

///OpenPartner
//-----------------------------------------------------------------------------

void TWindow::OpenPartner()
{
	if (ref.database < 0) return;
	RFile file(ref);
	char name[B_FILE_NAME_LENGTH+1];
	file.GetName(name);
	short nameLen = strlen(name);
	if (IsDotH(name)) {
		strcpy(name+nameLen-2, ".cpp");
	} else if (IsDotCpp(name)) {
		strcpy(name+nameLen-4, ".h");
	} else {
		return;
	}
	BDirectory dir;
	file.GetParent(&dir);
	if (dir.GetFile(name, &file) != B_NO_ERROR) return;
	BMessage *msg = new BMessage(B_REFS_RECEIVED);
	msg->AddRef("refs", file.Ref());
	be_app->PostMessage(msg);
}

///GetInfo
//-----------------------------------------------------------------------------

void TWindow::GetInfo()
{
	char str[2000];
	char *p = str;
	
	BTextView *textView = (BTextView*)FindView("textView");
	long len = textView->TextLength();
	
	sprintf(p, "%s\n\n", Title());
	p += strlen(p);
	
	long numLines = 0;
	for (const char *text = textView->Text(); *text; text++)
		if (*text == B_RETURN) numLines++;
	if (len > 0 && textView->CharAt(len-1) != B_RETURN) numLines++;
	if (numLines == 1) {
		sprintf(p, "1 line\n");
	} else {
		sprintf(p, "%ld lines\n", numLines);
	}
	p += strlen(p);
	
	if (len == 1) {
		sprintf(p, "1 character\n\n");
	} else {
		sprintf(p, "%ld characters\n\n", len);
	}
	p += strlen(p);
	
	if (ref.database < 0) {
		sprintf(p, "This file has not been saved.\n");
		p += strlen(p);
	} else {
		BFile file(ref);
		file.GetPath(p, str+1999-p);
		p += strlen(p);
		*p++ = B_RETURN;
		*p++ = B_RETURN;
		if (readonly) {
			sprintf(p, "This Be header file is read-only.");
		} else {
			sprintf(p, "This file has%s been modified since the last "
				"time is was saved.", dirty ? "" : " not");
		}
		p += strlen(p);
	}
	
	RAlert::InfoMsg(str, Frame());
}

///Ref
//-----------------------------------------------------------------------------

record_ref TWindow::Ref()
{
	record_ref result;
	Lock();
	result = ref;
	Unlock();
	return result;
}

///SetDirty
//-----------------------------------------------------------------------------

void TWindow::SetDirty(bool flag)
{
	Lock();
	if (flag != dirty) windows_menu_blooper->Check(this, flag);
	dirty = flag;
	if (!dirty) {
		TTextView *textView = (TTextView*)FindView("textView");
		textView->ClearUndo();
	}
	Unlock();
}

///Save
//-----------------------------------------------------------------------------

bool TWindow::Save(ulong what)
{
	// If this a Save command, if the window isn't dirty, don't
	// do anything. If it's dirty, and we have a file ref, save it.
	// Otherwise, fall through to the code to run the save panel.
	
	if (what == SAVE) {
		TTextView *textView = (TTextView*)FindView("textView");
		bool neverSavedButNotEmpty = ref.database < 0 && 
			textView->TextLength() > 0;
		if (!dirty && !neverSavedButNotEmpty) return TRUE;
		if (ref.database >= 0) {
			RFile file(ref);
			SaveTextToFile(file);
			SetDirty(FALSE);
			return TRUE;
		}
	}
	
	// Run the save panel.
	
	BMessage *msg = new BMessage(B_SAVE_REQUESTED);
	
	panel_info.Lock();
	if (panel_info.dir.database >= 0)
		msg->AddRef("directory", panel_info.dir);
	msg->AddRect("frame", panel_info.frame);
	panel_info.Unlock();
	
	if (what == SAVE_A_COPY) msg->AddBool("saveCopy", TRUE);
	char windowTitle[256];
	sprintf(windowTitle, 
		what == SAVE_A_COPY ? "BetterEdit: Save a copy of \"%s\" as ..." :
		"BetterEdit: Save \"%s\" as ...", 
		Title());
	RunSavePanel(Title(), windowTitle, NULL, msg);
	return FALSE;
}

///DisplayCompilerErrorMessage
//-----------------------------------------------------------------------------

void TWindow::DisplayCompilerErrorMessage(const char *emess1, 
	const char *emess2, const char *emess3)
{
	BView *panelView = FindView("panelView");
	if (panelView->IsHidden()) {
		BRect panelFrame = panelView->Frame();
		BScrollView *scrollView = (BScrollView*)FindView("scrollView");
		BRect scrollFrame = scrollView->Frame();
		long delta = panelFrame.bottom - panelFrame.top + 1;
		scrollView->MoveBy(0, delta);
		scrollView->ResizeBy(0, -delta);
		scrollView->Show();
		panelView->Show();
	}
	TPanelStringView *emess = (TPanelStringView*)FindView("emess1");
	emess->SetText(emess1);
	emess = (TPanelStringView*)FindView("emess2");
	emess->SetText(emess2);
	emess = (TPanelStringView*)FindView("emess3");
	emess->SetText(emess3);
}

///HidePanel
//-----------------------------------------------------------------------------

void TWindow::HidePanel()
{
	TPanelView *panelView = (TPanelView*)FindView("panelView");
	if (!panelView->IsHidden()) panelView->MouseDown(B_ORIGIN);
}
